/* ignuit - Educational software for the GNOME, following the Leitner
 * flash-card system.
 *
 * Copyright (C) 2008 Timothy Musson
 *
 * <trmusson@ihug.co.nz>
 * http://homepages.ihug.co.nz/~trmusson/programs.html#ignuit
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <config.h>
#include <gnome.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>

#include "main.h"
#include "prefs.h"
#include "app-window.h"
#include "dialog-properties.h"
#include "dialog-preferences.h"
#include "dialog-editor.h"
#include "dialog-quiz.h"
#include "dialog-find.h"
#include "dialog-about.h"
#include "fileio.h"


typedef struct {

    Ignuit           *ig;

    GladeXML         *gui;
    GtkWidget        *window;

    GtkWidget        *scrollw_categories;
    GtkTreeSelection *sel_categories;

    GtkWidget        *scrollw_cards;
    GtkTreeSelection *sel_cards;

    GtkWidget        *t_find;
    GtkWidget        *t_start_quiz;

    GtkWidget        *m_reset_stats;
    GtkWidget        *m_edit_card;
    GtkWidget        *m_add_card;
    GtkWidget        *m_remove_card;
    GtkWidget        *m_cut_card;
    GtkWidget        *m_copy_card;
    GtkWidget        *m_paste_card;
    GtkWidget        *m_select_all;
    GtkWidget        *m_find;
    GtkWidget        *m_find_all;
    GtkWidget        *m_view_trash;

    GtkWidget        *m_remove_category;

    GtkWidget        *m_switch_sides;

    GtkWidget        *m_start_quiz;
    GtkWidget        *m_quiz_reverse;

    GtkWidget        *r_quiz_category_selection;
    GSList           *quiz_category_group;

    GtkWidget        *r_quiz_card_selection;
    GSList           *quiz_card_group;

    GtkWidget        *b_remove_category;
    GtkWidget        *b_edit_card;
    GtkWidget        *b_add_card;
    GtkWidget        *b_remove_card;

    GtkWidget        *popup_menu_category;
    GtkWidget        *m_category_popup_remove;

    GtkWidget        *popup_menu_card;
    GtkWidget        *m_card_popup_add;
    GtkWidget        *m_card_popup_remove;
    GtkWidget        *m_card_popup_edit;
    GtkWidget        *m_card_popup_cut;
    GtkWidget        *m_card_popup_copy;
    GtkWidget        *m_card_popup_paste;
    GtkWidget        *m_card_popup_select_all;
    GtkWidget        *m_card_popup_switch_sides;
    GtkWidget        *m_card_popup_reset_stats;

    GtkWidget        *hpaned;

    GtkTreeViewColumn *treev_cat_title_column;
    GtkTreeViewColumn *treev_card_front_column;

    GList            *drag_list;

} WindowData;


/* Combobox ids for import/export filter selection: */
enum {
    FILTER_CSV = 0,
    FILTER_TSV,
    FILTER_NATIVE,
    FILTER_XSLT
};


/* Drag'n'drop */
static GtkTargetEntry target_table[] = {
    { "GTK_TREE_MODEL_ROW", GTK_TARGET_SAME_WIDGET, 0 },
    { "text/plain", GTK_TARGET_SAME_APP, 1 }
};


static void cb_category_row_inserted (GtkTreeModel *tree_model,
    GtkTreePath  *path,
    GtkTreeIter  *iter,
    WindowData   *data);


static gchar*
str_date (gchar *dest, GDate *date)
{
    /* Fill dest with the date in YYYY-MM-DD format, '\0' terminated.
     * The dest string must be already allocated to at least 11 chars.
     */

    if (date == NULL) {
        dest[0] = '-';
        dest[1] = '\0';
        return dest;
    }

    g_snprintf (dest, 11, "%04d-%02d-%02d",
        g_date_get_year (date),
        g_date_get_month (date),
        g_date_get_day (date));

    return dest;
}


static gchar*
truncated_card_text (const gchar *text)
{
    gchar *nl;

    /* Return a copy of the text, truncated at the first '\n' if present. */

    if ((nl = strchr (text, '\n')) == NULL) {
        return g_strdup (text);
    }

    return g_strndup (text, nl - text);
}


/*
static gchar*
truncated_card_front_text (Card *c)
{
    gchar *tr;

    tr = truncated_card_text (card_get_front_without_markup (c));

    return tr;
}
*/


static void
cell_func_card_expired (GtkTreeViewColumn *col,
    GtkCellRenderer *renderer,
    GtkTreeModel    *model,
    GtkTreeIter     *iter,
    gpointer        data)
{
    static gboolean expired;


    /* Indicate expired cards with an asterisk - otherwise blank. */

    gtk_tree_model_get (model, iter, COLUMN_CARD_EXPIRED, &expired, -1);
    if (expired) {

        /* FIXME: colour should be a preference option */
        g_object_set (renderer, "foreground", "#ff0000", NULL);
        g_object_set (renderer, "text", "*", NULL);

    }
    else {

        g_object_set (renderer, "text", "", NULL);

    }
}


static void
cell_func_card_group (GtkTreeViewColumn *col,
    GtkCellRenderer *renderer,
    GtkTreeModel    *model,
    GtkTreeIter     *iter,
    gpointer        data)
{
    static gchar s[2];
    static Group group;


    gtk_tree_model_get (model, iter, COLUMN_CARD_GROUP, &group, -1);

    s[0] = (group == GROUP_NEW) ? 'N' : '0' + group;
    s[1] = '\0';

    g_object_set (renderer, "text", s, NULL);
}


void
app_window_update_appbar (Ignuit *ig)
{
    Category *cat;
    gchar    *s;
    guint    nf, nc;


    nf = file_get_n_cards (ig->file);

    cat = file_get_current_category (ig->file);

    if (cat == NULL) {
        s = g_strdup_printf (_("File contains %d cards"), nf);
    }
    else if (file_category_is_trash (ig->file, cat)) {
        s = g_strdup_printf (_("Trash contains %d cards"),
            category_get_n_cards (cat));
    }
    else if (file_category_is_search (ig->file, cat)) {
        s = g_strdup_printf (_("Found %d cards"),
            category_get_n_cards (cat));
    }
    else {
        nc = category_get_n_cards (cat);
        s = g_strdup_printf (_("Category contains %d of %d cards"), nc, nf);
    }

    gnome_appbar_set_status (ig->appbar, s);

    g_free (s);
}


void
app_window_update_title (Ignuit *ig)
{
    const gchar *fname;
    gchar *s1, *s2;

    fname = file_get_filename (ig->file);
    if (fname) {
        s1 = g_filename_display_basename (fname);
    }
    else {
        s1 = g_strdup ("i GNU it");
    }

    if (file_get_changed (ig->file) == TRUE) {
        s2 = g_strdup_printf ("*%s", s1);
    }
    else {
        s2 = g_strdup (s1);
    }

    gtk_window_set_title (GTK_WINDOW(ig->app), s2);

    g_free (s1);
    g_free (s2);
}


static void
app_window_update_sensitivity (WindowData *data)
{
    Category *cat;
    gboolean file_has_cards;
    gboolean category_has_cards;
    gboolean can_paste;
    gboolean special;

    cat = file_get_current_category (data->ig->file);

    special = (cat == NULL ||
        file_category_is_search (data->ig->file, cat) ||
        file_category_is_trash (data->ig->file, cat) ||
        file_category_is_clipboard (data->ig->file, cat));

    file_has_cards = (file_get_n_cards (data->ig->file) > 0);
    category_has_cards = (cat && category_get_n_cards (cat) > 0);

    gtk_widget_set_sensitive (data->m_find_all, file_has_cards);
    gtk_widget_set_sensitive (data->t_start_quiz, file_has_cards);
    gtk_widget_set_sensitive (data->m_start_quiz, file_has_cards);
    gtk_widget_set_sensitive (data->b_add_card, !special);
    gtk_widget_set_sensitive (data->m_add_card, !special);
    gtk_widget_set_sensitive (data->m_card_popup_add, !special);
    gtk_widget_set_sensitive (data->m_view_trash,
        cat && file_get_trash (data->ig->file));
    gtk_widget_set_sensitive (data->b_remove_category, !special);
    gtk_widget_set_sensitive (data->m_remove_category, !special);
    gtk_widget_set_sensitive (data->m_category_popup_remove, !special);
    gtk_widget_set_sensitive (data->m_switch_sides, category_has_cards);
    gtk_widget_set_sensitive (data->m_card_popup_switch_sides,
        category_has_cards);
    gtk_widget_set_sensitive (data->m_reset_stats, category_has_cards);
    gtk_widget_set_sensitive (data->m_card_popup_reset_stats,
        category_has_cards);
    gtk_widget_set_sensitive (data->m_select_all, category_has_cards);
    gtk_widget_set_sensitive (data->m_card_popup_select_all,
        category_has_cards);
    gtk_widget_set_sensitive (data->m_find, file_has_cards);
    gtk_widget_set_sensitive (data->t_find, file_has_cards);

    can_paste = !special && (file_get_clipboard (data->ig->file) != NULL);

    gtk_widget_set_sensitive (data->m_paste_card, can_paste);
    gtk_widget_set_sensitive (data->m_card_popup_paste, can_paste);

    app_window_update_title (data->ig);
    app_window_update_appbar (data->ig);
}


static void
app_window_refresh_category_pane (Ignuit *ig)
{
    GList *cur;
    GtkTreeIter iter;
    Category *cat;
    GtkTreeModel *model;


    /* Clear and re-fill the category treeview. */

    model = gtk_tree_view_get_model (ig->treev_cat);
    gtk_list_store_clear (GTK_LIST_STORE(model));

    for (cur = file_get_categories(ig->file); cur != NULL; cur = cur->next) {

        cat = CATEGORY(cur);

        gtk_list_store_append (GTK_LIST_STORE(model), &iter);
        gtk_list_store_set (GTK_LIST_STORE(model), &iter,
            COLUMN_CATEGORY_DATA, cur,
            COLUMN_CATEGORY_TITLE, category_get_title (cat),
            COLUMN_CATEGORY_CARD_COUNT, category_get_n_cards (cat),
            -1);
    }
}


void
app_window_refresh_card_pane (Ignuit *ig, GList *cards)
{
    static gchar       crt[11], tst[11], exp[11];
    gchar              *front;
    static GtkTreeIter iter;
    GList              *cur;
    GtkTreeModel       *model;
    Card               *c;


    /* Clear and re-fill the card treeview. */


    /* For speed, temporarily detach model from view. */
    model = gtk_tree_view_get_model(ig->treev_card);
    g_object_ref(model);
    gtk_tree_view_set_model(ig->treev_card, NULL);

    gtk_list_store_clear (GTK_LIST_STORE(model));

    for (cur = cards; cur != NULL; cur = cur->next) {

        c = CARD(cur);

        crt[0] = '\0'; /* str_date (crt, card_get_gdate_created (c)); */
        str_date (tst, card_get_gdate_tested (c));
        str_date (exp, card_get_gdate_expiry (c));

        front = truncated_card_text (card_get_front_without_markup (c));

        gtk_list_store_append (GTK_LIST_STORE(model), &iter);
        gtk_list_store_set (GTK_LIST_STORE(model), &iter,
            COLUMN_CARD_DATA, cur,
            COLUMN_CARD_FRONT, front,
            COLUMN_CARD_BACK, NULL,
            COLUMN_CARD_GROUP, card_get_group (c),
            COLUMN_CARD_CREATED, crt,
            COLUMN_CARD_TESTED, tst,
            COLUMN_CARD_EXPIRED, card_get_expired (c),
            COLUMN_CARD_EXPIRY, exp,
            -1);

        g_free (front);
    }

    /* Re-attach model to view. */
    gtk_tree_view_set_model(ig->treev_card, model);
    g_object_unref(model);
}


GList*
app_window_get_category_list (Ignuit *ig)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    gboolean valid;
    GList *item, *list = NULL;

    /* Get a list of categories in the order displayed in the category pane. */

    model = gtk_tree_view_get_model (ig->treev_cat);
    valid = gtk_tree_model_get_iter_first (model, &iter);

    while (valid) {

        gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);
        list = g_list_append (list, item->data);
        valid = gtk_tree_model_iter_next (model, &iter);

    }
    return list;
}


GList*
app_window_find_treev_iter_with_data (GtkTreeView *treev,
    GtkTreeIter *iter,
    gpointer data)
{
    GtkTreeModel *model;
    gboolean valid;
    GList *item;


    /* Find the search data in the treeview, and set iter to it.
     * Return the found item, or NULL.
     */

    model = gtk_tree_view_get_model (treev);
    valid = gtk_tree_model_get_iter_first (model, iter);

    while (valid) {

        gtk_tree_model_get (model, iter, COLUMN_DATA, &item, -1);
        if (item->data == data) { return item; }
        valid = gtk_tree_model_iter_next (model, iter);

    }
    return NULL;
}


void
app_window_set_category_iter (Ignuit *ig, GtkTreeIter *iter, GList *item)
{
    GtkTreeModel *model;
    Category *cat;

    cat = CATEGORY(item);

    model = gtk_tree_view_get_model (ig->treev_cat);
    gtk_list_store_set (GTK_LIST_STORE(model), iter,
        COLUMN_CATEGORY_DATA, item,
        COLUMN_CATEGORY_TITLE, category_get_title (cat),
        COLUMN_CATEGORY_CARD_COUNT, category_get_n_cards (cat),
        -1);
}


void
app_window_refresh_category_row (Ignuit *ig, Category *cat)
{
    GtkTreeIter iter;
    GList *item;

    /* Make sure the row for this item is displaying current information. */

    item = app_window_find_treev_iter_with_data (ig->treev_cat, &iter, cat);
    g_assert (item != NULL);
    app_window_set_category_iter (ig, &iter, item);
}


void
app_window_set_card_iter (Ignuit *ig, GtkTreeIter *iter, GList *item)
{
    gchar        crt[11], tst[11], exp[11];
    gchar        *front;
    GtkTreeModel *model;
    Card         *c;


    /* Link this iter to the given item. */

    c = CARD(item);

    crt[0] = '\0'; /* str_date (crt, card_get_gdate_created (c)); */
    str_date (tst, card_get_gdate_tested (c));
    str_date (exp, card_get_gdate_expiry (c));

    front = truncated_card_text (card_get_front_without_markup (c));

    model = gtk_tree_view_get_model (ig->treev_card);
    gtk_list_store_set (GTK_LIST_STORE(model), iter,
        COLUMN_CARD_DATA, item,
        COLUMN_CARD_FRONT, front,
        COLUMN_CARD_BACK, NULL,
        COLUMN_CARD_GROUP, card_get_group (c),
        COLUMN_CARD_CREATED, crt,
        COLUMN_CARD_TESTED, tst,
        COLUMN_CARD_EXPIRED, card_get_expired (c),
        COLUMN_CARD_EXPIRY, exp,
        -1);

    g_free (front);
}


void
app_window_refresh_card_row (Ignuit *ig, GList *item)
{
    GtkTreeIter iter;
    gboolean ok;

    /* Make sure the row for this item is displaying current information. */

    ok = (app_window_find_treev_iter_with_data (ig->treev_card, &iter,
        CARD(item)) != NULL);
    g_assert (ok != FALSE);
    app_window_set_card_iter (ig, &iter, item);
}


void
app_window_add_card_iter (Ignuit *ig, GtkTreeIter *iter, GList *item)
{
    GtkTreeModel *model;
    GtkTreeIter sibling;


    /* Add new card to the treeview, and select it. */

    model = gtk_tree_view_get_model (ig->treev_card);
    if (gtk_tree_model_get_iter_first (model, &sibling)) {
        gtk_list_store_insert_before (GTK_LIST_STORE(model), iter, &sibling);
    }
    else {
        gtk_list_store_append (GTK_LIST_STORE(model), iter);
    }
    app_window_set_card_iter (ig, iter, item);
    app_window_select_card (ig, item);
}


void
app_window_select_treev_iter (GtkTreeView *treev, GtkTreeIter *iter)
{
    GtkTreeSelection *sel;
    GtkTreeModel *model;
    GtkTreePath *path;


    /* Select and scroll to a TreeView iter. */

    sel = gtk_tree_view_get_selection (treev);
    gtk_tree_selection_unselect_all (sel);

    model = gtk_tree_view_get_model (treev);
    path = gtk_tree_model_get_path (model, iter);

    gtk_tree_selection_select_iter (sel, iter);
    gtk_tree_view_scroll_to_cell (treev, path, NULL, TRUE, 0.5, 0.0);
}


gboolean
app_window_select_card (Ignuit *ig, GList *item)
{
    GtkTreeIter iter;
    gboolean ok;


    /* Select and scroll to an item in the card TreeView. */

    if (item == NULL) {
        GtkTreeSelection *sel;
        sel = gtk_tree_view_get_selection (ig->treev_card);
        gtk_tree_selection_unselect_all (sel);
        file_set_current_item (ig->file, NULL);
        return TRUE;
    }

    ok = (app_window_find_treev_iter_with_data
        (ig->treev_card, &iter, CARD(item)) != NULL);

    if (ok) {
        app_window_select_treev_iter (ig->treev_card, &iter);
    }

    return ok;
}


gboolean
app_window_select_category (Ignuit *ig, GList *item)
{
    GtkTreeIter iter;
    gboolean ok;


    /* Select and scroll to an item in the category TreeView. */

    if (item == NULL) {
        GtkTreeSelection *sel;
        sel = gtk_tree_view_get_selection (ig->treev_cat);
        gtk_tree_selection_unselect_all (sel);
        file_set_current_category (ig->file, NULL);
        return TRUE;
    }

    ok = (app_window_find_treev_iter_with_data
        (ig->treev_cat, &iter, CATEGORY(item)) != NULL);

    if (ok) {
        app_window_select_treev_iter (ig->treev_cat, &iter);
    }

    return ok;
}


gboolean
app_window_select_treev_root (GtkTreeView *treev)
{
    GtkTreeSelection *sel;
    GtkTreePath *path;
    gboolean ok;

    path = gtk_tree_path_new_first ();

    sel = gtk_tree_view_get_selection (treev);
    gtk_tree_selection_select_path (sel, path);

    ok = gtk_tree_selection_path_is_selected (sel, path);

    if (ok) {
        gtk_tree_view_scroll_to_cell (treev, path, NULL, TRUE, 0.5, 0.0);
    }

    gtk_tree_path_free (path);

    return ok;
}


static void
app_window_save_geometry (WindowData *data)
{
    gint w, h;

    /* Called just before exit to store window size, etc. */

    gtk_window_get_size (GTK_WINDOW(data->window), &w, &h);
    prefs_set_app_size (data->ig->prefs, w, h);

    w = gtk_paned_get_position (GTK_PANED(data->hpaned));
    prefs_set_category_pane_width (data->ig->prefs, w);

    w = gtk_tree_view_column_get_width (data->treev_cat_title_column);
    prefs_set_category_column_title_width (data->ig->prefs, w);

    w = gtk_tree_view_column_get_width (data->treev_card_front_column);
    prefs_set_card_column_front_width (data->ig->prefs, w);
}


static void
error_dialog (GtkWindow *parent, const gchar *message, GError *err)
{
    GtkDialog *dialog;
    gchar *msg;


    if (err && err->message) {
        msg = g_strdup_printf ("%s\n\n%s", message, err->message);
    }
    else {
        msg = g_strdup (message);
    }

    dialog = (GtkDialog*)gtk_message_dialog_new_with_markup (parent,
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_CLOSE,
        msg);

    gtk_dialog_run (dialog);
    gtk_widget_destroy (GTK_WIDGET(dialog));

    g_free (msg);
}


static gboolean
get_save_as_filename (WindowData *data)
{
    GtkWidget *dialog;
    gchar *fname_new;
    const gchar *fname_old;
    gboolean ret = FALSE;
    gint result;


    dialog = gtk_file_chooser_dialog_new (
        _("Save As"), GTK_WINDOW(data->window), GTK_FILE_CHOOSER_ACTION_SAVE,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
        NULL);

    gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

    fname_old = file_get_filename (data->ig->file);

    if (fname_old != NULL && g_path_is_absolute (fname_old)) {
        gtk_file_chooser_set_filename (GTK_FILE_CHOOSER(dialog), fname_old);
    }
    else if (prefs_get_workdir (data->ig->prefs)) {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            prefs_get_workdir (data->ig->prefs));
    }
    result = gtk_dialog_run (GTK_DIALOG(dialog));

    fname_new = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));

    if (result == GTK_RESPONSE_ACCEPT && fname_new != NULL) {

        if (g_file_test (fname_new, G_FILE_TEST_EXISTS)) {
            if (!ask_yes_or_no (GTK_WINDOW(data->window),
                _("<b>A file with the same name already exists</b>\n\nDo you want to replace it?"),
                GTK_RESPONSE_NO)) {
                result = GTK_RESPONSE_NONE;
            }
        }

        if (result == GTK_RESPONSE_ACCEPT) {
            file_set_filename (data->ig->file, fname_new);
            ret = TRUE;
        }

        g_free (fname_new);

    }

    gtk_widget_destroy (dialog);

    return ret;
}


static gint
ask_save (GtkWindow *parent)
{
    GtkDialog *dialog;
    gint result;

    dialog = (GtkDialog*)gtk_message_dialog_new_with_markup (parent,
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_QUESTION,
        GTK_BUTTONS_NONE,
        _("<b>The current file has not been saved</b>\n\nChanges since last save will be lost."));

    gtk_dialog_add_buttons (dialog,
        _("Close _without Saving"), GTK_RESPONSE_REJECT,
        GTK_STOCK_CANCEL,           GTK_RESPONSE_CANCEL,
        GTK_STOCK_SAVE,             GTK_RESPONSE_ACCEPT,
        NULL);

    gtk_dialog_set_default_response (dialog, GTK_RESPONSE_ACCEPT);
    result = gtk_dialog_run (dialog);
    gtk_widget_destroy (GTK_WIDGET(dialog));

    if (result != GTK_RESPONSE_ACCEPT && result != GTK_RESPONSE_REJECT) {
        result = GTK_RESPONSE_CANCEL;
    }

    return result;
}


gboolean
ask_yes_or_no (GtkWindow *parent, const gchar *question,
    GtkResponseType def)
{
    GtkDialog *dialog;
    gint result;

    dialog = (GtkDialog*)gtk_message_dialog_new_with_markup (parent,
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_QUESTION,
        GTK_BUTTONS_YES_NO,
        question);

    gtk_dialog_set_default_response (dialog, def);
    result = gtk_dialog_run (dialog);
    gtk_widget_destroy (GTK_WIDGET(dialog));

    return result == GTK_RESPONSE_YES;
}


static void
unable_to_save_file (GtkWindow *parent, GError *err)
{
    error_dialog (parent, _("<b>Unable to save file</b>"), err);

     if (err != NULL) {
        g_error_free (err);
    }
}


static gboolean
save_changes (WindowData *data)
{
    GError *err = NULL;
    gint reply;


    /* If the file has been changed, ask whether or not to save it.
     * Return TRUE only if the changes are either unwanted or
     * successfully saved.
     */

    if (!file_get_changed (data->ig->file)) {
        return TRUE;
    }

    reply = ask_save (GTK_WINDOW(data->window));

    if (reply == GTK_RESPONSE_CANCEL) {
        return FALSE;
    }

    if (reply == GTK_RESPONSE_REJECT) {
        return TRUE;
    }

    if (!file_get_filename (data->ig->file) && !get_save_as_filename (data)) {
        return FALSE;
    }

    if (!fileio_save (data->ig, file_get_filename (data->ig->file), &err)) {

        unable_to_save_file (GTK_WINDOW(data->window), err);

        return FALSE;
    }

    app_window_update_title (data->ig);

    return TRUE;
}


static void
cb_m_new (GtkWidget *widget, WindowData *data)
{
    GList *first;

    /* Close the current file and create a new one. */

    dialog_editor_check_changed ();

    if (!save_changes (data)) { return; }

    dialog_editor_kill ();

    file_free (data->ig->file, TRUE);

    data->ig->file = file_new ();
    first = file_add_category (data->ig->file, category_new (NULL));
    file_set_current_category (data->ig->file, CATEGORY(first));

    app_window_refresh_category_pane (data->ig);
    app_window_refresh_card_pane (data->ig,
        file_get_current_category_cards (data->ig->file));

    app_window_update_sensitivity (data);
}


static void
cb_m_open (GtkWidget *widget, WindowData *data)
{
    GtkWidget *dialog;
    File *file;
    gint result;
    gchar *fname;

    /* Prompt for a file to open, and try to load it. If successful, close
     * the current file and display the newly loaded one.
     */

    dialog_editor_check_changed ();

    if (!save_changes (data)) { return; }

    dialog_editor_kill ();

    dialog = gtk_file_chooser_dialog_new (_("Open"),
        GTK_WINDOW(data->window), GTK_FILE_CHOOSER_ACTION_OPEN,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
        NULL);
    gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

    if (prefs_get_workdir (data->ig->prefs)) {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            prefs_get_workdir (data->ig->prefs));
    }

    result = gtk_dialog_run (GTK_DIALOG(dialog));
    fname = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    gtk_widget_destroy (dialog);

    if (result == GTK_RESPONSE_ACCEPT && fname != NULL) {

        GError *err = NULL;

        if ((file = fileio_load (fname, &err)) != NULL) {

            /* Success - replace the current file with the opened one. */

            GtkTreeModel *model;

            file_set_current_category (file, NULL);
            file_free (data->ig->file, TRUE);
            data->ig->file = file;

            if (g_path_is_absolute (fname)) {
                gchar *workdir;
                workdir = g_path_get_dirname (fname);
                prefs_set_workdir (data->ig->prefs, workdir);
                g_free (workdir);
            }

            model = gtk_tree_view_get_model (data->ig->treev_cat);

            g_signal_handlers_block_by_func (G_OBJECT(model),
                G_CALLBACK(cb_category_row_inserted), data);

            app_window_refresh_category_pane (data->ig);

            g_signal_handlers_unblock_by_func (G_OBJECT(model),
                G_CALLBACK(cb_category_row_inserted), data);

            if (file_get_categories (data->ig->file)) {
                app_window_select_treev_root (data->ig->treev_cat);
            }
            else {
                app_window_update_sensitivity (data);
                app_window_refresh_card_pane (data->ig,
                    file_get_current_category_cards (data->ig->file));
            }

            app_window_update_title (data->ig);

        }
        else {

            error_dialog (GTK_WINDOW(data->window),
                _("<b>Unable to load file</b>"), err);

            if (err != NULL) {
                g_error_free (err);
            }

        }
        g_free (fname);
    }
}


static void
cb_m_save (GtkWidget *widget, WindowData *data)
{
    GError *err = NULL;

    dialog_editor_check_changed ();

    if (file_get_filename (data->ig->file) == NULL) {
        if (!get_save_as_filename (data)) {
            return;
        }
    }

    if (!fileio_save (data->ig, file_get_filename (data->ig->file), &err)) {
        unable_to_save_file (GTK_WINDOW(data->window), err);
        return;
    }

    app_window_update_title (data->ig);
}


static void
cb_m_save_as (GtkWidget *widget, WindowData *data)
{
    dialog_editor_check_changed ();

    if (get_save_as_filename (data)) {

        GError *err = NULL;

        if (!fileio_save (data->ig, file_get_filename (data->ig->file), &err)) {

            unable_to_save_file (GTK_WINDOW(data->window), err);
            return;

        }
        app_window_update_title (data->ig);

    }
}


static GList*
populate_import_export_combo_box (WindowData *data, GtkComboBox *combo,
    gboolean is_import)
{
    GList       *list = NULL;
    GDir        *dir;
    const gchar *whichdir;
    gchar       *dname;

    whichdir = is_import ? IMPORT_DIR : EXPORT_DIR;

    dname = gnome_program_locate_file (data->ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, whichdir, TRUE, NULL);

    if (!dname) {
        g_warning ("Can't find directory: %s\n", whichdir);
        return NULL;
    }

    gtk_combo_box_append_text (combo, _("CSV"));
    gtk_combo_box_append_text (combo, _("TSV"));
    gtk_combo_box_append_text (combo, "Ignuit");

    dir = g_dir_open (dname, 0, NULL);
    if (dir) {

        const gchar *entry;

        while ((entry = g_dir_read_name (dir))) {

            gchar *s, *ext;

            s = g_filename_to_utf8 (entry, -1, NULL, NULL, NULL);
            if (s != NULL) {
                if ((ext = strchr (s, '.')) != NULL) {
                    *ext = '\0';
                }
                gtk_combo_box_append_text (combo, s);
                g_free (s);
                list = g_list_append (list, g_strdup (entry));
            }
            else {
                g_warning ("g_filename_to_utf8 failed for '%s'\n", entry);
            }
        }

        g_dir_close (dir);
    }
    g_free (dname);

    return list;
}


static void
cb_m_import (GtkWidget *widget, WindowData *data)
{
    GtkWidget *dialog;
    GtkWidget *box;
    GtkWidget *combo;
    GtkWidget *label;
    GList     *filters;
    File      *file;
    gint      result, combo_selection;
    gchar     *fname;


    /* Import a non-native file, adding it as a new category to the
     * currently open file.
     */

    dialog_editor_check_changed ();
    dialog_editor_kill (); /* XXX: maybe nicer not to */

    dialog = gtk_file_chooser_dialog_new (_("Import"),
        GTK_WINDOW(data->window), GTK_FILE_CHOOSER_ACTION_OPEN,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
        NULL);
    gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

    if (prefs_get_workdir (data->ig->prefs)) {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            prefs_get_workdir (data->ig->prefs));
    }

    box = gtk_hbox_new (FALSE, 0);

    combo = gtk_combo_box_new_text ();
    filters = populate_import_export_combo_box (data,
        GTK_COMBO_BOX(combo), TRUE);
    gtk_combo_box_set_active (GTK_COMBO_BOX(combo), 0);

    gtk_box_pack_end (GTK_BOX(box), combo, FALSE, FALSE, 0);

    label = gtk_label_new_with_mnemonic (_("Import _Filter:"));
    gtk_box_pack_end (GTK_BOX(box), label, FALSE, FALSE, 12);

    gtk_file_chooser_set_extra_widget (GTK_FILE_CHOOSER(dialog), box);
    gtk_widget_show_all (box);

    result = gtk_dialog_run (GTK_DIALOG(dialog));
    fname = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    combo_selection = gtk_combo_box_get_active (GTK_COMBO_BOX(combo));
    gtk_widget_destroy (dialog);

    if (result == GTK_RESPONSE_ACCEPT && fname != NULL) {

        GError *err = NULL;

        const gchar *fname_filter;

        switch (combo_selection) {
        case FILTER_CSV:
            file = fileio_import_sv (fname, ",", &err);
            break;
        case FILTER_TSV:
            file = fileio_import_sv (fname, "\t", &err);
            break;
        case FILTER_NATIVE:
            file = fileio_load (fname, &err);
            break;
        default:
            /* >= FILTER_XSLT */
            fname_filter = g_list_nth_data (filters,
                combo_selection - FILTER_XSLT);
            file = fileio_import (data->ig, fname, fname_filter, &err);
            break;
        }

        if (file) {

            Category *cat, *cat_tmp;
            GList *cur, *item;
            GtkTreeIter iter;

            if (g_path_is_absolute (fname)) {
                gchar *workdir;
                workdir = g_path_get_dirname (fname);
                prefs_set_workdir (data->ig->prefs, workdir);
                g_free (workdir);
            }

            /* Add cards from the imported file to the currently open file. */

            cat_tmp = file_get_current_category (data->ig->file);

            for (cur = file_get_categories (file); cur != NULL; cur = cur->next) {

                GList *c;

                cat = category_new (category_get_title (CATEGORY(cur)));

                file_add_category (data->ig->file, cat);
                file_set_current_category (data->ig->file, cat);

                for (c = category_get_cards (CATEGORY(cur)); c != NULL; c = c->next) {
                    file_add_loaded_card (data->ig->file, CARD(c));
                }

            }

            file_set_changed (data->ig->file, TRUE);
            file_set_current_category (data->ig->file, cat_tmp);

            app_window_refresh_category_pane (data->ig);
            app_window_refresh_card_pane (data->ig, category_get_cards (cat));
            app_window_update_appbar (data->ig);

            file_free (file, FALSE);

            item = app_window_find_treev_iter_with_data
                (data->ig->treev_cat, &iter, cat_tmp);
            app_window_select_category (data->ig, item);

        }
        else {

            error_dialog (GTK_WINDOW(data->window),
                _("<b>Unable to import file</b>"), err);

            if (err != NULL) {
                g_error_free (err);
            }

        }

    }

    if (filters) {
        g_list_foreach (filters, (GFunc)g_free, NULL);
        g_list_free (filters);
    }
}


static void
cb_m_export (GtkWidget *widget, WindowData *data)
{
    GtkWidget *dialog;
    GtkWidget *box;
    GtkWidget *combo;
    GtkWidget *label;
    GList     *filters;
    gchar     *fname;
    gint      result, combo_selection;


    dialog = gtk_file_chooser_dialog_new (
        _("Export"), GTK_WINDOW(data->window), GTK_FILE_CHOOSER_ACTION_SAVE,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
        NULL);

    gtk_dialog_set_default_response (GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

    if (prefs_get_workdir (data->ig->prefs)) {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),
            prefs_get_workdir (data->ig->prefs));
    }

    box = gtk_hbox_new (FALSE, 0);

    combo = gtk_combo_box_new_text ();
    filters = populate_import_export_combo_box (data,
        GTK_COMBO_BOX(combo), FALSE);
    gtk_combo_box_set_active (GTK_COMBO_BOX(combo), 0);

    gtk_box_pack_end (GTK_BOX(box), combo, FALSE, FALSE, 0);

    label = gtk_label_new_with_mnemonic (_("Export _Filter:"));
    gtk_box_pack_end (GTK_BOX(box), label, FALSE, FALSE, 12);

    gtk_file_chooser_set_extra_widget (GTK_FILE_CHOOSER(dialog), box);
    gtk_widget_show_all (box);

    result = gtk_dialog_run (GTK_DIALOG(dialog));
    fname = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    combo_selection = gtk_combo_box_get_active (GTK_COMBO_BOX(combo));
    gtk_widget_destroy (dialog);

    if (result == GTK_RESPONSE_ACCEPT && fname != NULL) {

        GError *err = NULL;

        const gchar *filter;
        gboolean ok, changed_tmp;

        switch (combo_selection) {
        case FILTER_CSV:
            ok = fileio_export_sv (data->ig->file, fname, ",", &err);
            break;
        case FILTER_TSV:
            ok = fileio_export_sv (data->ig->file, fname, "\t", &err);
            break;
        case FILTER_NATIVE:
            changed_tmp = file_get_changed (data->ig->file);
            ok = fileio_save (data->ig, fname, &err);
            file_set_changed (data->ig->file, changed_tmp);
            break;
        default:
            /* >= FILTER_XSLT */
            filter = g_list_nth_data (filters,
                combo_selection - FILTER_XSLT);
            ok = fileio_export (data->ig, fname, filter, &err);
            break;
        }

        if (!ok) {

            error_dialog (GTK_WINDOW(data->window),
                _("<b>Unable to export file</b>"), err);

            if (err != NULL) {
                g_error_free (err);
            }

        }

        g_free (fname);

    }

    if (filters) {
        g_list_foreach (filters, (GFunc)g_free, NULL);
        g_list_free (filters);
    }
}


static void
cb_m_properties (GtkWidget *widget, WindowData *data)
{
    dialog_properties (data->ig);
}


static void
cb_m_quit (GtkWidget *widget, WindowData *data)
{
    dialog_editor_check_changed ();

    if (!save_changes (data)) {return;}

    app_window_save_geometry (data);

    g_object_unref (G_OBJECT(data->gui));
    g_free (data);

    gtk_main_quit ();
}


static void
cb_m_add_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeIter iter;
    GList *item;
    Category *cat;
    Card *c;


    /* Make a new blank card, add it to the currently selected category,
     * and open the card editor to fill it in.
     */

    dialog_editor_check_changed ();

    cat = file_get_current_category (data->ig->file);

    c = card_new ();
    card_set_front (c, "");
    card_set_back (c, "");
    file_add_card (data->ig->file, cat, c);
    file_set_changed (data->ig->file, TRUE);

    item = category_get_cards (cat);
    file_set_current_item (data->ig->file, item);

    app_window_add_card_iter (data->ig, &iter, item);
    app_window_refresh_category_row (data->ig, cat);

    app_window_update_sensitivity (data);

    dialog_editor (data->ig);
}


static void
cb_m_remove_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeRowReference *rowref;
    GList *rows, *rowref_list, *cur, *item;
    Category *cat_cur;
    Card *c;


    /* Move any currently selected cards to the trash. */

    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);
    if (rows == NULL) {
        return;
    }

    dialog_editor_check_changed ();

    rowref_list = NULL;

    for (cur = rows; cur != NULL; cur = cur->next) {
        path = (GtkTreePath*)cur->data;
        rowref = gtk_tree_row_reference_new (model, path);
        rowref_list = g_list_append (rowref_list, rowref);
    }

    cat_cur = file_get_current_category (data->ig->file);

    for (cur = rowref_list; cur != NULL; cur = cur->next) {

        if ((path = gtk_tree_row_reference_get_path
            ((GtkTreeRowReference*)cur->data)) != NULL) {

            if (gtk_tree_model_get_iter (model, &iter, path)) {

                Category *cat;

                gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
                gtk_list_store_remove (GTK_LIST_STORE(model), &iter);

                c = CARD(item);
                cat = card_get_category (c);

                /* Remove it, but don't free it just yet. */
                if (file_category_is_search (data->ig->file, cat_cur)) {
                    file_remove_card (data->ig->file, c);
                    category_remove_card (cat_cur, c);
                }
                else {
                    file_remove_card (data->ig->file, c);
                }

                if (file_category_is_trash (data->ig->file, cat_cur)
                    || card_is_blank (c)) {
                    /* Might as well delete it immediately. */
                    card_free (c);
                }
                else {
                    /* Move it to the trash. */
                    file_add_trash (data->ig->file, c);
                }

                if (!file_category_is_trash (data->ig->file, cat)) {
                    app_window_refresh_category_row (data->ig, cat);
                }

                file_set_changed (data->ig->file, TRUE);

            }

            gtk_tree_path_free (path);
        }
    }

    g_list_foreach (rowref_list, (GFunc)gtk_tree_row_reference_free, NULL);
    g_list_free (rowref_list);

    g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (rows);

    if (gtk_list_store_iter_is_valid (GTK_LIST_STORE(model), &iter)) {
        gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
        gtk_tree_selection_select_iter (data->sel_cards, &iter);
        file_set_current_item (data->ig->file, item);
    }
    else {
        file_set_current_item (data->ig->file, category_get_cards (cat_cur));
        app_window_select_card (data->ig, category_get_cards (cat_cur));
    }

    app_window_update_sensitivity (data);

    dialog_editor_tweak ();
}


static void
cb_m_edit_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GList *rows, *item;


    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);
    g_assert (rows != NULL);

    path = (GtkTreePath*)rows->data; /* Use the first selected row. */

    gtk_tree_model_get_iter (model, &iter, path);
    gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
    app_window_select_treev_iter (data->ig->treev_card, &iter);

    file_set_current_item (data->ig->file, item);

    g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (rows);

    dialog_editor (data->ig);
}


static void
cb_m_find (GtkWidget *widget, WindowData *data)
{
    dialog_find (data->ig);
}


static void
cb_m_find_all (GtkWidget *widget, WindowData *data)
{
    Category *cat;
    GList *cur, *cards;

    dialog_editor_check_changed ();

    file_clear_search (data->ig->file, TRUE);

    for (cur = file_get_cards (data->ig->file); cur != NULL; cur = cur->next) {

        file_add_search_card (data->ig->file, CARD(cur));

    }

    cat = file_get_search (data->ig->file);
    cards = category_get_cards (cat);

    app_window_refresh_card_pane (data->ig, cards);
    app_window_select_category (data->ig, NULL);
    app_window_select_card (data->ig, cards);

    file_set_current_category (data->ig->file, cat);
    file_set_current_item (data->ig->file, cards);

    app_window_update_sensitivity (data);
    app_window_update_appbar (data->ig);

    dialog_editor_tweak ();
}


static void
cb_m_view_trash (GtkWidget *widget, WindowData *data)
{
    Category *trash;
    GList *cards;


    dialog_editor_check_changed ();

    trash = file_get_trash (data->ig->file);
    cards = category_get_cards (trash);

    app_window_refresh_card_pane (data->ig, cards);
    app_window_select_category (data->ig, NULL);
    app_window_select_card (data->ig, cards);

    file_set_current_category (data->ig->file, trash);
    file_set_current_item (data->ig->file, cards);

    app_window_update_sensitivity (data);
    app_window_update_appbar (data->ig);

    dialog_editor_tweak ();
}


static void
cb_m_add_category (GtkWidget *widget, WindowData *data)
{
    GtkTreeIter iter;
    GtkTreeModel *model;
    Category *cat;
    GList *item;


    /* Create a new category and add it to the file. */

    cat = category_new (NULL);
    item = file_add_category (data->ig->file, cat);
    file_set_changed (data->ig->file, TRUE);

    /* Update the category TreeView. Note that this emits row-inserted on
     * the tree model, which results in cb_category_inserted being called. */

    model = gtk_tree_view_get_model (data->ig->treev_cat);
    gtk_tree_model_get_iter_first (model, &iter);
    gtk_list_store_append (GTK_LIST_STORE(model), &iter);
    gtk_list_store_set (GTK_LIST_STORE(model), &iter,
        COLUMN_CATEGORY_DATA, item,
        COLUMN_CATEGORY_TITLE, category_get_title (cat),
        -1);

    /* Switch to the new category. */

    file_set_current_category (data->ig->file, cat);
    file_set_current_item (data->ig->file, NULL);

    app_window_refresh_card_pane (data->ig, NULL);
    app_window_update_sensitivity (data);
    dialog_editor_tweak ();
}


static void
cb_m_remove_category (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GList *item, *next;
    Category *cat;


    /* Delete the selected category (and any cards it contains). */

    if (gtk_tree_selection_get_selected (data->sel_categories, &model,
        &iter)) {

        gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);
        if (item == NULL) { return; }

        cat = CATEGORY(item);
        if (category_get_n_cards (cat) && !ask_yes_or_no
            (GTK_WINDOW(data->window),
            _("<b>Delete this non-empty category?</b>\n\nThe category and its contents will be permanently deleted."),
            GTK_RESPONSE_NO)) {
            return;
        }

        dialog_editor_check_changed ();

        file_remove_category (data->ig->file, cat);

        if (gtk_list_store_remove (GTK_LIST_STORE(model), &iter)) {
            gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &next, -1);
            gtk_tree_selection_select_iter (data->sel_categories, &iter);
        }
        else {
            if (!app_window_select_treev_root (data->ig->treev_cat)) {
                /* We've just deleted the last available category. */
                file_set_current_category (data->ig->file, NULL);
                file_set_current_item (data->ig->file, NULL);
                app_window_refresh_card_pane (data->ig, NULL);
                dialog_editor_tweak ();
            }
        }

        app_window_update_sensitivity (data);
    }
}


static void
cb_m_select_all (GtkWidget *widget, WindowData *data)
{
    GtkTreeSelection *sel;

    sel = gtk_tree_view_get_selection (data->ig->treev_card);
    gtk_tree_selection_select_all (sel);
}


static void
cb_m_switch_sides (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeRowReference *rowref;
    GList *rows, *rowref_list, *cur, *item;


    /* Swap the front and back text of each selected card. */

    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);
    if (rows == NULL) {
        return;
    }

    dialog_editor_check_changed ();

    rowref_list = NULL;

    for (cur = rows; cur != NULL; cur = cur->next) {
        path = (GtkTreePath*)cur->data;
        rowref = gtk_tree_row_reference_new (model, path);
        rowref_list = g_list_append (rowref_list, rowref);
    }

    for (cur = rowref_list; cur != NULL; cur = cur->next) {

        if ((path = gtk_tree_row_reference_get_path
            ((GtkTreeRowReference*)cur->data)) != NULL) {

            if (gtk_tree_model_get_iter (model, &iter, path)) {

                gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);

                card_switch_sides (CARD(item));

                app_window_set_card_iter (data->ig, &iter, item);

            }

            gtk_tree_path_free (path);
        }
    }

    g_list_foreach (rowref_list, (GFunc)gtk_tree_row_reference_free, NULL);
    g_list_free (rowref_list);

    g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (rows);

    file_set_changed (data->ig->file, TRUE);
    app_window_update_title (data->ig);

    dialog_editor_tweak ();
}


static void
cb_m_reset_stats (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeRowReference *rowref;
    GList *rows, *rowref_list, *cur, *item;


    /* Forget the dates and counts for all selected cards. */

    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);
    if (rows == NULL) {
        return;
    }

    dialog_editor_check_changed ();

    if (!ask_yes_or_no (GTK_WINDOW(data->window),
        _("<b>Expiry dates will be reset</b>\n\nRecorded dates and statistics for the selected cards will be lost. Continue anyway?"),
        GTK_RESPONSE_NO)) {
        return;
    }

    rowref_list = NULL;

    for (cur = rows; cur != NULL; cur = cur->next) {
        path = (GtkTreePath*)cur->data;
        rowref = gtk_tree_row_reference_new (model, path);
        rowref_list = g_list_append (rowref_list, rowref);
    }

    for (cur = rowref_list; cur != NULL; cur = cur->next) {

        if ((path = gtk_tree_row_reference_get_path
            ((GtkTreeRowReference*)cur->data)) != NULL) {

            if (gtk_tree_model_get_iter (model, &iter, path)) {

                gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);

                card_reset_statistics (CARD(item));

                app_window_set_card_iter (data->ig, &iter, item);

            }

            gtk_tree_path_free (path);
        }
    }

    g_list_foreach (rowref_list, (GFunc)gtk_tree_row_reference_free, NULL);
    g_list_free (rowref_list);

    g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (rows);

    file_set_changed (data->ig->file, TRUE);
    app_window_update_title (data->ig);
}


static void
cb_m_paste_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeIter iter;
    GList *cur, *item;
    Category *cat, *clipboard;
    Card *c;


    /* Move cards from the clipboard to the currently selected category. */

    clipboard = file_get_clipboard (data->ig->file);
    g_assert (clipboard != NULL);

    dialog_editor_check_changed ();

    cat = file_get_current_category (data->ig->file);

    while ((cur = category_get_cards (clipboard))) {

        c = CARD(cur);
        category_remove_card (clipboard, c);
        file_add_card (data->ig->file, cat, c);

        item = category_get_cards (cat);
        file_set_current_item (data->ig->file, item);
        app_window_add_card_iter (data->ig, &iter, item);
        app_window_refresh_category_row (data->ig, cat);

    }

    file_set_changed (data->ig->file, TRUE);
    file_clear_clipboard (data->ig->file);
    app_window_update_sensitivity (data);

    dialog_editor_tweak ();
}


static void
cb_m_copy_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeRowReference *rowref;
    GList *rows, *rowref_list, *cur, *item;
    Card *copy;


    /* Duplicate any currently selected cards, storing the duplicates
     * in the clipboard. */

    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);
    if (rows == NULL) {
        return;
    }

    dialog_editor_check_changed ();

    file_clear_clipboard (data->ig->file);

    rowref_list = NULL;

    for (cur = rows; cur != NULL; cur = cur->next) {
        path = (GtkTreePath*)cur->data;
        rowref = gtk_tree_row_reference_new (model, path);
        rowref_list = g_list_append (rowref_list, rowref);
    }

    for (cur = rowref_list; cur != NULL; cur = cur->next) {

        if ((path = gtk_tree_row_reference_get_path
            ((GtkTreeRowReference*)cur->data)) != NULL) {

            if (gtk_tree_model_get_iter (model, &iter, path)) {

                gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);

                copy = card_copy (CARD(item));
                file_add_clipboard (data->ig->file, copy);

            }

            gtk_tree_path_free (path);
        }
    }

    g_list_foreach (rowref_list, (GFunc)gtk_tree_row_reference_free, NULL);
    g_list_free (rowref_list);

    g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (rows);

    if (file_get_clipboard (data->ig->file) != NULL) {
        gtk_widget_set_sensitive (data->m_paste_card, TRUE);
        gtk_widget_set_sensitive (data->m_card_popup_paste, TRUE);
    }
}


static void
cb_m_cut_card (GtkWidget *widget, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeRowReference *rowref;
    GList *rows, *rowref_list, *cur, *item;
    Category *cat_cur;
    Card *c;


    /* Move any currently selected cards to the clipboard. */

    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);
    if (rows == NULL) {
        return;
    }

    dialog_editor_check_changed ();

    file_clear_clipboard (data->ig->file);

    rowref_list = NULL;

    for (cur = rows; cur != NULL; cur = cur->next) {
        path = (GtkTreePath*)cur->data;
        rowref = gtk_tree_row_reference_new (model, path);
        rowref_list = g_list_append (rowref_list, rowref);
    }

    cat_cur = file_get_current_category (data->ig->file);

    for (cur = rowref_list; cur != NULL; cur = cur->next) {

        if ((path = gtk_tree_row_reference_get_path
            ((GtkTreeRowReference*)cur->data)) != NULL) {

            if (gtk_tree_model_get_iter (model, &iter, path)) {

                Category *cat;

                gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
                gtk_list_store_remove (GTK_LIST_STORE(model), &iter);

                c = CARD(item);
                cat = card_get_category (c);

                /* Remove the card from the file. */
                if (file_category_is_search (data->ig->file, cat_cur)) {
                    file_remove_card (data->ig->file, c);
                    category_remove_card (cat_cur, c);
                }
                else {
                    file_remove_card (data->ig->file, c);
                }

                file_add_clipboard (data->ig->file, c);

                if (!file_category_is_trash (data->ig->file, cat)) {
                    app_window_refresh_category_row (data->ig, cat);
                }

                file_set_changed (data->ig->file, TRUE);

            }

            gtk_tree_path_free (path);
        }
    }

    g_list_foreach (rowref_list, (GFunc)gtk_tree_row_reference_free, NULL);
    g_list_free (rowref_list);

    g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (rows);

    if (gtk_list_store_iter_is_valid (GTK_LIST_STORE(model), &iter)) {
        gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
        gtk_tree_selection_select_iter (data->sel_cards, &iter);
        file_set_current_item (data->ig->file, item);
    }
    else {
        file_set_current_item (data->ig->file, category_get_cards (cat_cur));
        app_window_select_card (data->ig, category_get_cards (cat_cur));
    }

    if (file_get_clipboard (data->ig->file) != NULL) {
        gtk_widget_set_sensitive (data->m_paste_card, TRUE);
        gtk_widget_set_sensitive (data->m_card_popup_paste, TRUE);
    }

    app_window_update_sensitivity (data);

    dialog_editor_tweak ();
}


static void
cb_m_preferences (GtkWidget *widget, WindowData *data)
{
    dialog_preferences (data->ig);
}


static void
cb_m_help (GtkWidget *widget, WindowData *data)
{
    GError *err = NULL;

    gnome_help_display ("ignuit.xml", NULL, &err);

    if (err != NULL) {
        error_dialog (GTK_WINDOW(data->window),
            _("Couldn't display user guide"), err);
        g_error_free (err);
    }
}


static gint
get_quiz_category_selection (WindowData *data)
{
    GSList *cur;
    gint i = QUIZ_N_CATEGORY_SELECTIONS;

    for (cur = data->quiz_category_group; cur != NULL; cur = cur->next) {

        i--;

        if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM(cur->data))) {
            return i;
        }
    }
    g_assert_not_reached ();
}


static gint
get_quiz_card_selection (WindowData *data)
{
    GSList *cur;
    gint i = QUIZ_N_CARD_SELECTIONS;

    for (cur = data->quiz_card_group; cur != NULL; cur = cur->next) {

        i--;

        if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM(cur->data))) {
            return i;
        }
    }
    g_assert_not_reached ();
}


static void
cb_m_start_quiz (GtkWidget *widget, WindowData *data)
{
    dialog_editor_check_changed ();
    dialog_editor_kill ();

    data->ig->quizinfo.category_selection = get_quiz_category_selection (data);
    data->ig->quizinfo.card_selection = get_quiz_card_selection (data);
    data->ig->quizinfo.reverse = gtk_check_menu_item_get_active
        (GTK_CHECK_MENU_ITEM(data->m_quiz_reverse));

    dialog_quiz (data->ig);
}


static void
cb_m_about (GtkWidget *widget, WindowData *data)
{
    dialog_about (data->window);
}


static gboolean
cb_category_pane_popup (GtkWidget *widget, GdkEventButton *event,
    WindowData *data)
{
    if (event->type != GDK_BUTTON_PRESS || event->button != 3) {
        return FALSE;
    }

    gtk_menu_popup (GTK_MENU(data->popup_menu_category),
        NULL, NULL, NULL, NULL, 3, event->time);

    return TRUE;
}


static gboolean
cb_card_pane_popup (GtkWidget *widget, GdkEventButton *event,
    WindowData *data)
{
    if (event->type != GDK_BUTTON_PRESS || event->button != 3) {
        return FALSE;
    }

    gtk_menu_popup (GTK_MENU(data->popup_menu_card),
        NULL, NULL, NULL, NULL, 3, event->time);

    return TRUE;
}


static void
tweak_editor (WindowData *data)
{
    if (file_get_current_category_cards (data->ig->file)) {
        app_window_select_treev_root (data->ig->treev_card);
    }
    dialog_editor_tweak ();
}


static void
cb_treev_category_selection (GtkTreeSelection *sel, WindowData *data)
{
    GtkTreeIter iter;
    GtkTreeModel *model;


    /* A category has been selected in the category TreeView.
     * Update the card TreeView accordingly.
     */

    if (gtk_tree_selection_get_selected (sel, &model, &iter)) {

        GtkAdjustment *adj;
        GList *item;
        Category *cat;

        dialog_editor_check_changed ();

        gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);
        cat = item ? CATEGORY(item) : NULL;

        if (cat == NULL) {
            return;
        }

        file_set_current_category (data->ig->file, cat);
        file_set_current_item (data->ig->file,
            cat ? category_get_cards (cat) : NULL);

        adj = gtk_scrolled_window_get_vadjustment
            (GTK_SCROLLED_WINDOW(data->scrollw_cards));
        gtk_adjustment_set_value (adj, 0);

        app_window_refresh_card_pane (data->ig,
            file_get_current_category_cards (data->ig->file));

        tweak_editor (data);

        app_window_update_sensitivity (data);
    }
}


static void
cb_category_row_inserted (GtkTreeModel *tree_model,
    GtkTreePath  *path,
    GtkTreeIter  *iter,
    WindowData   *data)
{
    GtkTreeView *treev;
    GtkTreeSelection *sel;


    /* Select the newly added category. Called after adding a new category
     * or drag'n'dropping an existing row.
     */

    treev = data->ig->treev_cat;

    sel = gtk_tree_view_get_selection (treev);
    gtk_tree_selection_select_path (sel, path);

    gtk_tree_view_scroll_to_cell (treev, path, NULL, TRUE, 0.5, 0.0);
}


static void
cb_treev_category_title_edited (GtkCellRendererText *cell,
    gchar *path_string, gchar *new_text, WindowData *data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;


    /* A category's title has been edited in the category TreeView. */

    model = gtk_tree_view_get_model (data->ig->treev_cat);
    path = gtk_tree_path_new_from_string (path_string);

    if (gtk_tree_model_get_iter (model, &iter, path)) {

        GList *item;

        gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);

        category_set_title (CATEGORY(item), new_text);

        gtk_list_store_set (GTK_LIST_STORE(model), &iter,
            COLUMN_CATEGORY_TITLE, new_text, -1);

        file_set_changed (data->ig->file, TRUE);
    }
    gtk_tree_path_free (path);
}


static void
cb_treev_card_activated (GtkTreeView *treev, GtkTreePath *path,
    GtkTreeViewColumn *column, WindowData *data)
{
    GtkTreeIter iter;
    GtkTreeModel *model;

    /* A row in the card TreeView has been double-clicked. */

    model = gtk_tree_view_get_model (treev);
    if (gtk_tree_model_get_iter (model, &iter, path)) {

        /* Make this the only selected card. */

        GList *item;

        dialog_editor_check_changed ();

        gtk_tree_selection_unselect_all (data->sel_cards);
        gtk_tree_selection_select_iter (data->sel_cards, &iter);
        gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
        file_set_current_item (data->ig->file, item);

        /* Edit the card. */

        dialog_editor (data->ig);
    }
}


static gboolean
drag_selected_foreach_func (GtkTreeModel  *model,
    GtkTreePath *path, GtkTreeIter *iter, WindowData *data)
{
    GList *item;

    gtk_tree_model_get (model, iter, COLUMN_CARD_DATA, &item, -1);
    data->drag_list = g_list_prepend (data->drag_list, CARD(item));
    return TRUE;
}


static void
cb_treev_cards_drag_data_get (GtkTreeView *treev,
    GdkDragContext *context, GtkSelectionData *sln_data, guint info,
    guint t, WindowData *data)
{
    GtkTreeModel *model;
    GList *rows;


    /* XXX
     * It works, but I don't think it's the correct way to do drag'n'drop.
     */

    rows = gtk_tree_selection_get_selected_rows (data->sel_cards, &model);

    if (rows != NULL) {

        gtk_tree_selection_selected_foreach (data->sel_cards,
            (GtkTreeSelectionForeachFunc)drag_selected_foreach_func, data);

        g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
        g_list_free (rows);
    }
}


static void
cb_treev_categories_drag_data_received (GtkTreeView *treev,
    GdkDragContext *context,
    gint x, gint y,
    GtkSelectionData *sln_data,
    guint info,
    guint time,
    WindowData *data)
{
    GList *cur, *item, *rowref_list = NULL;
    GtkTreeRowReference *rowref;
    GtkTreePath *path;
    GtkTreeModel *model;
    GtkTreeIter iter;
    Category *cat_dest, *cat_cur;


    /* XXX
     * It works, but I don't think it's the correct way to do drag'n'drop.
     */

    /* If this is just a category row-reordering, ignore it. */
    if (gtk_drag_get_source_widget (context) ==
        GTK_WIDGET(data->ig->treev_cat)) {
        file_set_changed (data->ig->file, TRUE);
        return;
    }

    /* Get drag destination row. */
    gtk_tree_view_get_dest_row_at_pos (treev, x, y, &path, NULL);
    model = gtk_tree_view_get_model (treev);
    gtk_tree_model_get_iter (model, &iter, path);

    /* Get drag destination category. */
    gtk_tree_model_get (model, &iter, COLUMN_CATEGORY_DATA, &item, -1);
    cat_dest = CATEGORY(item);

    /* Build a list of row references from the dragged cards. */
    model = gtk_tree_view_get_model (data->ig->treev_card);
    for (cur = data->drag_list; cur != NULL; cur = cur->next) {

        app_window_find_treev_iter_with_data (data->ig->treev_card, &iter,
            CARD(cur));

        path = gtk_tree_model_get_path (model, &iter);
        rowref = gtk_tree_row_reference_new (model, path);
        rowref_list = g_list_append (rowref_list, rowref);

        gtk_tree_path_free (path);
    }


    dialog_editor_check_changed ();


    /* Move the cards to their new category. */

    /* Where the cards are begin dragged from. */
    cat_cur = file_get_current_category (data->ig->file);

    for (cur = rowref_list; cur != NULL; cur = cur->next) {

        if ((path = gtk_tree_row_reference_get_path
            ((GtkTreeRowReference*)cur->data)) != NULL) {

            if (gtk_tree_model_get_iter (model, &iter, path)) {

                Card *c;
                Category *cat_src;

                gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);

                c = CARD(item);
                cat_src = card_get_category (c);

                if (cat_src == cat_dest) {
                /* XXX: Should remove then add the card back to the treev. */
                    continue;
                }

                gtk_list_store_remove (GTK_LIST_STORE(model), &iter);

                /* Unlink the card from the file. */
                if (file_category_is_search (data->ig->file, cat_cur)) {
                    file_remove_card (data->ig->file, c);
                    category_remove_card (cat_cur, c);
                }
                else {
                    file_remove_card (data->ig->file, c);
                }

                /* Add the card to its new category. */
                file_add_card (data->ig->file, cat_dest, c);

                if (!file_category_is_trash (data->ig->file, cat_cur)) {
                    app_window_refresh_category_row (data->ig, cat_src);
                }

                file_set_changed (data->ig->file, TRUE);
            }

            gtk_tree_path_free (path);
        }
    }

    g_list_foreach (rowref_list, (GFunc)gtk_tree_row_reference_free, NULL);
    g_list_free (rowref_list);

    g_list_free (data->drag_list);
    data->drag_list = NULL;

    if (gtk_list_store_iter_is_valid (GTK_LIST_STORE(model), &iter)) {
        gtk_tree_model_get (model, &iter, COLUMN_CARD_DATA, &item, -1);
        gtk_tree_selection_select_iter (data->sel_cards, &iter);
        file_set_current_item (data->ig->file, item);
    }
    else {
        file_set_current_item (data->ig->file, category_get_cards (cat_cur));
    }

    app_window_refresh_category_row (data->ig, cat_dest);
    app_window_update_sensitivity (data);

    dialog_editor_tweak ();

    gtk_drag_finish (context, FALSE, FALSE, time);
}


static void
cb_treev_cards_selection (GtkTreeSelection *sel, WindowData *data)
{
    GtkTreeModel *model;
    GList *rows;
    gboolean selected;


    rows = gtk_tree_selection_get_selected_rows (sel, &model);
    selected = (rows != NULL);

    gtk_widget_set_sensitive (data->m_edit_card, selected);
    gtk_widget_set_sensitive (data->m_remove_card, selected);
    gtk_widget_set_sensitive (data->m_cut_card, selected);
    gtk_widget_set_sensitive (data->m_copy_card, selected);

    gtk_widget_set_sensitive (data->m_card_popup_edit, selected);
    gtk_widget_set_sensitive (data->m_card_popup_remove, selected);
    gtk_widget_set_sensitive (data->m_card_popup_cut, selected);
    gtk_widget_set_sensitive (data->m_card_popup_copy, selected);

    gtk_widget_set_sensitive (data->b_edit_card, selected);
    gtk_widget_set_sensitive (data->b_remove_card, selected);

    if (rows) {

        g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
        g_list_free (rows);
    }
}


static void
cb_window_destroy (GtkWidget *widget, WindowData *data)
{
    g_object_unref (G_OBJECT(data->gui));
    g_free (data);

    gtk_main_quit ();
}


static gboolean
cb_window_delete (GtkWidget *widget, GdkEvent *event, WindowData *data)
{
    dialog_editor_check_changed ();

    if (!save_changes (data)) {return TRUE;}

    app_window_save_geometry (data);

    return FALSE;
}


static void
add_category_column (WindowData *data, const gchar *title, gint col,
    GtkTreeCellDataFunc func)
{
    GtkCellRenderer *r;
    GtkTreeViewColumn *c;

    /* Add a column to the category TreeView. */

    r = gtk_cell_renderer_text_new ();
    c = gtk_tree_view_column_new_with_attributes (title, r, "text", col, NULL);
    gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN(c), TRUE);
    gtk_tree_view_append_column (data->ig->treev_cat, c);
    if (func != NULL) {
        gtk_tree_view_column_set_cell_data_func (c, r, func, NULL, NULL);
    }

    if (col == COLUMN_CATEGORY_TITLE) {
        gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN(c),
            GTK_TREE_VIEW_COLUMN_FIXED);
        g_object_set (r, "ellipsize", PANGO_ELLIPSIZE_END, NULL);
        g_object_set (r, "editable", TRUE, NULL);

        g_signal_connect(r, "edited",
            (GCallback)cb_treev_category_title_edited, data);

        data->treev_cat_title_column = c;
    }
}


static void
add_card_column (WindowData *data, const gchar *title, gint col,
    GtkTreeCellDataFunc func)
{
    GtkCellRenderer *r;
    GtkTreeViewColumn *c;


    /* Add a column to the card TreeView. */

    r = gtk_cell_renderer_text_new ();
    c = gtk_tree_view_column_new_with_attributes (title, r, "text", col, NULL);
    gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN(c), TRUE);
    gtk_tree_view_append_column (data->ig->treev_card, c);
    if (func != NULL) {
        gtk_tree_view_column_set_cell_data_func (c, r, func, NULL, NULL);
    }

    if (col == COLUMN_CARD_FRONT) {
        gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN(c),
            GTK_TREE_VIEW_COLUMN_FIXED);
        g_object_set (r, "ellipsize", PANGO_ELLIPSIZE_END, NULL);
        data->treev_card_front_column = c;
    }
}


void
app_window (Ignuit *ig)
{
    WindowData *data;
    GtkWidget *m_new, *m_open, *m_save_as, *m_quit;
    GtkWidget *m_import, *m_export;
    GtkWidget *m_add_category, *b_add_category, *m_category_popup_add;
    GtkWidget *m_properties, *m_preferences;
    GtkWidget *m_help, *m_about;
    GtkWidget *t_new, *t_open, *t_help;
    GtkTreeModel *model;
    GtkListStore *list_store;
    gchar *imgfile, *gladefile, *iconfile;


    gladefile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, F_GLADE_MAIN, TRUE, NULL);

    if (gladefile == NULL) {
        g_warning ("Can't find file: %s\n", F_GLADE_MAIN);
        exit (EXIT_FAILURE);
    }

    data = g_new0 (WindowData, 1);
    data->gui = glade_xml_new (gladefile, NULL, NULL);
    g_free (gladefile);

    data->ig = ig;

    data->drag_list = NULL;

    ig->app = glade_xml_get_widget (data->gui, "app");

    data->window = ig->app;

    gtk_window_set_default_size (GTK_WINDOW(data->window),
        prefs_get_app_width (data->ig->prefs),
        prefs_get_app_height (data->ig->prefs));

    m_new = glade_xml_get_widget (data->gui, "m_new");
    m_open = glade_xml_get_widget (data->gui, "m_open");
    data->ig->m_save = glade_xml_get_widget (data->gui, "m_save");
    m_save_as = glade_xml_get_widget (data->gui, "m_save_as");
    m_import = glade_xml_get_widget (data->gui, "m_import");
    m_export = glade_xml_get_widget (data->gui, "m_export");
    data->m_reset_stats = glade_xml_get_widget (data->gui, "m_reset_stats");
    m_properties = glade_xml_get_widget (data->gui, "m_properties");
    m_quit = glade_xml_get_widget (data->gui, "m_quit");

    data->m_add_card = glade_xml_get_widget (data->gui, "m_add_card");
    data->m_remove_card = glade_xml_get_widget (data->gui, "m_remove_card");
    data->m_edit_card = glade_xml_get_widget (data->gui, "m_edit_card");
    data->m_cut_card = glade_xml_get_widget (data->gui, "m_cut_card");
    data->m_copy_card = glade_xml_get_widget (data->gui, "m_copy_card");
    data->m_paste_card = glade_xml_get_widget (data->gui, "m_paste_card");
    data->m_select_all = glade_xml_get_widget (data->gui, "m_select_all");
    data->m_find = glade_xml_get_widget (data->gui, "m_find");
    data->m_find_all = glade_xml_get_widget (data->gui, "m_show_all");
    data->m_view_trash = glade_xml_get_widget (data->gui, "m_view_trash");
    m_add_category = glade_xml_get_widget (data->gui, "m_add_category");
    data->m_remove_category = glade_xml_get_widget (data->gui, "m_remove_category");
    data->m_switch_sides = glade_xml_get_widget (data->gui, "m_switch_sides");
    m_preferences = glade_xml_get_widget (data->gui, "m_preferences");

    data->m_start_quiz = glade_xml_get_widget (data->gui, "m_start_quiz");

    data->r_quiz_category_selection = glade_xml_get_widget (data->gui,
        "r_all_categories");

    data->quiz_category_group = gtk_radio_menu_item_get_group
        (GTK_RADIO_MENU_ITEM(data->r_quiz_category_selection));

    data->r_quiz_card_selection = glade_xml_get_widget
        (data->gui, "r_all_cards");

    data->quiz_card_group = gtk_radio_menu_item_get_group
        (GTK_RADIO_MENU_ITEM(data->r_quiz_card_selection));

    data->m_quiz_reverse = glade_xml_get_widget (data->gui, "m_quiz_reverse");

    m_help = glade_xml_get_widget (data->gui, "m_help");
    m_about = glade_xml_get_widget (data->gui, "m_about");

    t_new = glade_xml_get_widget (data->gui, "t_new");
    t_open = glade_xml_get_widget (data->gui, "t_open");
    data->ig->t_save = glade_xml_get_widget (data->gui, "t_save");
    data->t_find = glade_xml_get_widget (data->gui, "t_find");
    data->t_start_quiz = glade_xml_get_widget (data->gui, "t_start_quiz");
    t_help = glade_xml_get_widget (data->gui, "t_help");

    if ((imgfile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, F_IMG_BTN_QUIZ_1, TRUE, NULL))) {
        GtkWidget *img = gtk_image_new_from_file (imgfile);
        gtk_tool_button_set_icon_widget
            (GTK_TOOL_BUTTON(data->t_start_quiz), img);
        g_free (imgfile);
        gtk_widget_show (img);
    }

    b_add_category = glade_xml_get_widget (data->gui, "b_add_category");
    data->b_remove_category = glade_xml_get_widget
        (data->gui, "b_remove_category");

    data->b_edit_card = glade_xml_get_widget (data->gui, "b_edit_card");
    data->b_add_card = glade_xml_get_widget (data->gui, "b_add_card");
    data->b_remove_card = glade_xml_get_widget (data->gui, "b_remove_card");

    data->hpaned = glade_xml_get_widget (data->gui, "hpaned");

    ig->appbar = (GnomeAppBar*)glade_xml_get_widget (data->gui, "appbar");



    /* category list */

    data->scrollw_categories = glade_xml_get_widget (data->gui,
        "scrollw_categories");
    ig->treev_cat = GTK_TREE_VIEW(glade_xml_get_widget (data->gui,
        "treeview_categories"));

    list_store = gtk_list_store_new (CATEGORIES_N_COLUMNS,
        G_TYPE_POINTER,
        G_TYPE_STRING,
        G_TYPE_INT);

    gtk_tree_view_set_model (data->ig->treev_cat,
        GTK_TREE_MODEL(list_store));

    g_object_unref (list_store);

    add_category_column (data, _("Category"), COLUMN_CATEGORY_TITLE, NULL);
    add_category_column (data, NULL, COLUMN_CATEGORY_CARD_COUNT, NULL);

    data->sel_categories = gtk_tree_view_get_selection (data->ig->treev_cat);
    gtk_tree_selection_set_mode (data->sel_categories, GTK_SELECTION_BROWSE);
    gtk_tree_view_set_reorderable (data->ig->treev_cat, TRUE);

    g_signal_connect (G_OBJECT(data->sel_categories), "changed",
        G_CALLBACK (cb_treev_category_selection), data);


    /* category pane right-click menu */

    data->popup_menu_category = glade_xml_get_widget (data->gui,
        "popup_menu_category");

    m_category_popup_add = glade_xml_get_widget (data->gui,
        "m_category_popup_add");
    g_signal_connect (G_OBJECT(m_category_popup_add), "activate",
        G_CALLBACK(cb_m_add_category), data);

    data->m_category_popup_remove = glade_xml_get_widget (data->gui,
        "m_category_popup_remove");
    g_signal_connect (G_OBJECT(data->m_category_popup_remove),
        "activate", G_CALLBACK(cb_m_remove_category), data);


    /* card pane right-click menu */

    data->popup_menu_card = glade_xml_get_widget (data->gui,
        "popup_menu_card");

    data->m_card_popup_add = glade_xml_get_widget (data->gui,
        "m_card_popup_add");
    g_signal_connect (G_OBJECT(data->m_card_popup_add), "activate",
        G_CALLBACK(cb_m_add_card), data);

    data->m_card_popup_remove = glade_xml_get_widget (data->gui,
        "m_card_popup_remove");
    g_signal_connect (G_OBJECT(data->m_card_popup_remove), "activate",
        G_CALLBACK(cb_m_remove_card), data);

    data->m_card_popup_edit = glade_xml_get_widget (data->gui,
        "m_card_popup_edit");
    g_signal_connect (G_OBJECT(data->m_card_popup_edit), "activate",
        G_CALLBACK(cb_m_edit_card), data);

    data->m_card_popup_cut = glade_xml_get_widget (data->gui,
        "m_card_popup_cut");
    g_signal_connect (G_OBJECT(data->m_card_popup_cut), "activate",
        G_CALLBACK(cb_m_cut_card), data);
    data->m_card_popup_copy = glade_xml_get_widget (data->gui,
        "m_card_popup_copy");
    g_signal_connect (G_OBJECT(data->m_card_popup_copy), "activate",
        G_CALLBACK(cb_m_copy_card), data);
    data->m_card_popup_paste = glade_xml_get_widget (data->gui,
        "m_card_popup_paste");
    g_signal_connect (G_OBJECT(data->m_card_popup_paste), "activate",
        G_CALLBACK(cb_m_paste_card), data);

    data->m_card_popup_select_all = glade_xml_get_widget (data->gui,
        "m_card_popup_select_all");
    g_signal_connect (G_OBJECT(data->m_card_popup_select_all), "activate",
        G_CALLBACK(cb_m_select_all), data);

    data->m_card_popup_switch_sides = glade_xml_get_widget (data->gui,
        "m_card_popup_switch_sides");
    g_signal_connect (G_OBJECT(data->m_card_popup_switch_sides), "activate",
        G_CALLBACK(cb_m_switch_sides), data);

    data->m_card_popup_reset_stats = glade_xml_get_widget (data->gui,
        "m_card_popup_reset_stats");
    g_signal_connect (G_OBJECT(data->m_card_popup_reset_stats), "activate",
        G_CALLBACK(cb_m_reset_stats), data);


    /* card list */

    data->scrollw_cards = glade_xml_get_widget (data->gui, "scrollw_cards");
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(data->scrollw_cards),
        GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
    ig->treev_card = GTK_TREE_VIEW(glade_xml_get_widget (data->gui,
        "treeview_cards"));

    /* data->treev_cards = ig->treev_card; */

    list_store = gtk_list_store_new (CARDS_N_COLUMNS,
        G_TYPE_POINTER, /* this card's category GList node */
        G_TYPE_STRING,  /* front */
        G_TYPE_STRING,  /* back */
        G_TYPE_INT,     /* group */
        G_TYPE_STRING,  /* created */
        G_TYPE_STRING,  /* date last tested */
        G_TYPE_BOOLEAN, /* expired */
        G_TYPE_STRING); /* date of expiry */

    gtk_tree_view_set_model (data->ig->treev_card, GTK_TREE_MODEL(list_store));
    g_object_unref (list_store);

    add_card_column (data, _("Front"), COLUMN_CARD_FRONT, NULL);
    add_card_column (data, _("B") /* Box */ , COLUMN_CARD_GROUP,
        cell_func_card_group);
    add_card_column (data, _("Tested"), COLUMN_CARD_TESTED, NULL);
    add_card_column (data, _("!") /* Expired! */ , COLUMN_CARD_EXPIRED,
        cell_func_card_expired);
    add_card_column (data, _("Expiry"), COLUMN_CARD_EXPIRY, NULL);

    data->sel_cards = gtk_tree_view_get_selection (data->ig->treev_card);
    gtk_tree_selection_set_mode (data->sel_cards, GTK_SELECTION_MULTIPLE);

    g_signal_connect (G_OBJECT(data->sel_cards), "changed",
        G_CALLBACK (cb_treev_cards_selection), data);


    /* signals */

    g_signal_connect (G_OBJECT(data->window), "destroy",
        G_CALLBACK(cb_window_destroy), data);
    g_signal_connect (G_OBJECT(data->window), "delete_event",
        G_CALLBACK(cb_window_delete), data);

    g_signal_connect (G_OBJECT(m_new), "activate",
        G_CALLBACK(cb_m_new), data);
    g_signal_connect (G_OBJECT(m_open), "activate",
        G_CALLBACK(cb_m_open), data);
    g_signal_connect (G_OBJECT(data->ig->m_save), "activate",
        G_CALLBACK(cb_m_save), data);
    g_signal_connect (G_OBJECT(m_save_as), "activate",
        G_CALLBACK(cb_m_save_as), data);
    g_signal_connect (G_OBJECT(m_import), "activate",
        G_CALLBACK(cb_m_import), data);
    g_signal_connect (G_OBJECT(m_export), "activate",
        G_CALLBACK(cb_m_export), data);
    g_signal_connect (G_OBJECT(data->m_reset_stats), "activate",
        G_CALLBACK(cb_m_reset_stats), data);
    g_signal_connect (G_OBJECT(m_properties), "activate",
        G_CALLBACK(cb_m_properties), data);
    g_signal_connect (G_OBJECT(m_quit), "activate",
        G_CALLBACK(cb_m_quit), data);

    g_signal_connect (G_OBJECT(data->m_edit_card), "activate",
        G_CALLBACK(cb_m_edit_card), data);
    g_signal_connect (G_OBJECT(data->m_cut_card), "activate",
        G_CALLBACK(cb_m_cut_card), data);
    g_signal_connect (G_OBJECT(data->m_copy_card), "activate",
        G_CALLBACK(cb_m_copy_card), data);
    g_signal_connect (G_OBJECT(data->m_paste_card), "activate",
        G_CALLBACK(cb_m_paste_card), data);
    g_signal_connect (G_OBJECT(data->m_find), "activate",
        G_CALLBACK(cb_m_find), data);
    g_signal_connect (G_OBJECT(data->m_find_all), "activate",
        G_CALLBACK(cb_m_find_all), data);
    g_signal_connect (G_OBJECT(data->m_view_trash), "activate",
        G_CALLBACK(cb_m_view_trash), data);
    g_signal_connect (G_OBJECT(data->m_add_card), "activate",
        G_CALLBACK(cb_m_add_card), data);
    g_signal_connect (G_OBJECT(data->m_remove_card), "activate",
        G_CALLBACK(cb_m_remove_card), data);
    g_signal_connect (G_OBJECT(m_add_category), "activate",
        G_CALLBACK(cb_m_add_category), data);
    g_signal_connect (G_OBJECT(data->m_remove_category), "activate",
        G_CALLBACK(cb_m_remove_category), data);
    g_signal_connect (G_OBJECT(data->m_select_all), "activate",
        G_CALLBACK(cb_m_select_all), data);
    g_signal_connect (G_OBJECT(data->m_switch_sides), "activate",
        G_CALLBACK(cb_m_switch_sides), data);
    g_signal_connect (G_OBJECT(m_preferences), "activate",
        G_CALLBACK(cb_m_preferences), data);

    g_signal_connect (G_OBJECT(data->m_start_quiz), "activate",
        G_CALLBACK(cb_m_start_quiz), data);

    g_signal_connect (G_OBJECT(m_help), "activate",
        G_CALLBACK(cb_m_help), data);
    g_signal_connect (G_OBJECT(m_about), "activate",
        G_CALLBACK(cb_m_about), data);

    g_signal_connect (G_OBJECT(t_new), "clicked",
        G_CALLBACK(cb_m_new), data);
    g_signal_connect (G_OBJECT(t_open), "clicked",
        G_CALLBACK(cb_m_open), data);
    g_signal_connect (G_OBJECT(data->ig->t_save), "clicked",
        G_CALLBACK(cb_m_save), data);
    g_signal_connect (G_OBJECT(data->t_find), "clicked",
        G_CALLBACK(cb_m_find), data);
    g_signal_connect (G_OBJECT(data->t_start_quiz), "clicked",
        G_CALLBACK(cb_m_start_quiz), data);
    g_signal_connect (G_OBJECT(t_help), "clicked",
        G_CALLBACK(cb_m_help), data);

    g_signal_connect (G_OBJECT(b_add_category), "clicked",
        G_CALLBACK(cb_m_add_category), data);
    g_signal_connect (G_OBJECT(data->b_remove_category), "clicked",
        G_CALLBACK(cb_m_remove_category), data);

    g_signal_connect (G_OBJECT(data->b_edit_card), "clicked",
        G_CALLBACK(cb_m_edit_card), data);
    g_signal_connect (G_OBJECT(data->b_add_card), "clicked",
        G_CALLBACK(cb_m_add_card), data);
    g_signal_connect (G_OBJECT(data->b_remove_card), "clicked",
        G_CALLBACK(cb_m_remove_card), data);

    model = gtk_tree_view_get_model (data->ig->treev_cat);
    g_signal_connect (G_OBJECT(model), "row-inserted",
        G_CALLBACK(cb_category_row_inserted), data);

    g_signal_connect (G_OBJECT(data->ig->treev_cat), "button-press-event",
        G_CALLBACK(cb_category_pane_popup), data);
    g_signal_connect (G_OBJECT(data->ig->treev_card), "button-press-event",
        G_CALLBACK(cb_card_pane_popup), data);

    g_signal_connect (G_OBJECT(data->ig->treev_card), "row-activated",
        G_CALLBACK(cb_treev_card_activated), data);



    /* Drag'n'drop */

    gtk_tree_view_enable_model_drag_source (data->ig->treev_cat,
        GDK_BUTTON1_MASK,
        target_table,
        G_N_ELEMENTS(target_table),
        GDK_ACTION_MOVE);

    gtk_tree_view_enable_model_drag_dest (data->ig->treev_cat,
        target_table,
        G_N_ELEMENTS(target_table),
        GDK_ACTION_MOVE);

    gtk_tree_view_enable_model_drag_source (data->ig->treev_card,
        GDK_BUTTON1_MASK,
        target_table,
        G_N_ELEMENTS(target_table),
        GDK_ACTION_MOVE);

    g_signal_connect (G_OBJECT(data->ig->treev_card), "drag_data_get",
        G_CALLBACK(cb_treev_cards_drag_data_get), data);
    g_signal_connect (G_OBJECT(data->ig->treev_cat), "drag_data_received",
        G_CALLBACK(cb_treev_categories_drag_data_received), data);

    iconfile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_PIXMAP, F_ICON, FALSE, NULL);

    if (iconfile != NULL) {
        gtk_window_set_icon_from_file (GTK_WINDOW(data->window),
            iconfile, NULL);
        g_free (iconfile);
    }

    app_window_refresh_category_pane (data->ig);
    app_window_refresh_card_pane (data->ig,
        file_get_current_category_cards (data->ig->file));

    gtk_paned_set_position (GTK_PANED(data->hpaned),
        prefs_get_category_pane_width (data->ig->prefs));

    gtk_tree_view_column_set_fixed_width (data->treev_cat_title_column,
        prefs_get_category_column_title_width (data->ig->prefs));

    gtk_tree_view_column_set_fixed_width (data->treev_card_front_column,
        prefs_get_card_column_front_width (data->ig->prefs));

    app_window_update_sensitivity (data);

    /* Scroll to first category */
    app_window_select_treev_root (data->ig->treev_cat);


    ig->m_remove_category = data->m_remove_category;
    ig->b_remove_category = data->b_remove_category;
    ig->m_add_card = data->m_add_card;
    ig->b_add_card = data->b_add_card;
    ig->m_start_quiz = data->m_start_quiz;
    ig->t_start_quiz = data->t_start_quiz;
    ig->m_find = data->m_find;
    ig->t_find = data->t_find;
    ig->m_find_all = data->m_find_all;
    ig->m_view_trash = data->m_view_trash;
    ig->m_switch_sides = data->m_switch_sides;
    ig->m_reset_stats = data->m_reset_stats;
    ig->m_paste_card = data->m_paste_card;
    ig->m_select_all = data->m_select_all;

    ig->m_card_popup_paste = data->m_card_popup_paste;
    ig->m_card_popup_switch_sides = data->m_card_popup_switch_sides;
    ig->m_card_popup_reset_stats = data->m_card_popup_reset_stats;
    ig->m_card_popup_select_all = data->m_card_popup_select_all;

    gtk_widget_show (data->window);
}

