/* ignuit - Educational software for the GNOME, following the Leitner
 * flash-card system.
 *
 * Copyright (C) 2008 Timothy Musson
 *
 * <trmusson@ihug.co.nz>
 * http://homepages.ihug.co.nz/~trmusson/programs.html#ignuit
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <config.h>
#include <gnome.h>
#include <glade/glade.h>

#include "main.h"
#include "file.h"
#include "prefs.h"
#include "dialog-quiz.h"
#include "app-window.h"


#define CUR_IS_LAST(d)          ((d)->cur->data == (d)->last)
#define QUIZITEM(glist_item)    ((QuizItem*)(glist_item)->data)
#define QCARD(q)                (QUIZITEM(q)->card)


/* XXX: this belongs in card.h */
void card_mark_tested (Prefs *p, Card *c, gboolean known);


typedef struct _QuizItem QuizItem;

struct _QuizItem {

    Card *card;

    gboolean tested;
    gboolean known;

};


typedef struct {

    Ignuit              *ig;

    GladeXML            *gui;
    GtkWidget           *window;

    GtkTextView         *textview;
    GtkTextBuffer       *textbuf;

    GList               *cardview_selected;

    GtkWidget           *m_prev;
    GtkWidget           *m_next;
    GtkCheckMenuItem    *m_known;
    GtkCheckMenuItem    *m_unknown;
    GtkCheckMenuItem    *m_flip;
    GtkWidget           *b_prev;
    GtkWidget           *b_next;
    GtkWidget           *chk_store;
    GtkToggleButton     *b_known;
    GtkToggleButton     *b_unknown;
    GtkToggleButton     *b_flip;

    GtkStatusbar        *statusbar;
    guint               sbcid;

    gboolean            ugly_toggle_hack;

    GList               *quizitems;
    GList               *cur;

    QuizItem            *last;      /* The final "End of quiz" message. */

    gint                n_quizitems;
    gint                n;

} DialogData;


typedef enum {
    TOKEN_NONE = 0,
    TOKEN_TEXT,
    TOKEN_NEW_LINE,
    TOKEN_TAG
} Token;


typedef enum {
    ELEMENT_ID_UNKNOWN = 0,
    ELEMENT_ID_KEYWORD,
    ELEMENT_ID_BOLD,
    ELEMENT_ID_ITALIC,
    ELEMENT_ID_UNDERLINE,
    ELEMENT_ID_STRIKETHROUGH,
    ELEMENT_ID_BIG,
    ELEMENT_ID_SMALL,
    ELEMENT_ID_SUB,
    ELEMENT_ID_SUP,
    ELEMENT_ID_CENTER
} ElementID;


typedef struct _Element Element;


struct _Element {
    gchar     *name;
    gint      length;
};


static Element element[] = {
    { NULL,     0 },
    { "key",    3 },
    { "b",      1 },
    { "i",      1 },
    { "u",      1 },
    { "s",      1 },
    { "big",    3 },
    { "small",  5 },
    { "sub",    3 },
    { "sup",    3 },
    { "center", 6 }
};


static gint n_elements = sizeof (element) / sizeof (element[0]);


typedef struct _Parser Parser;

struct _Parser {

    GString   *data;
    Token     token;
    gint      length;
    ElementID element_id;
    gboolean  is_open_tag;
    gboolean  in_keyword;
    gboolean  error;
    GList     *tags;

};


typedef struct _Tag Tag;

struct _Tag {

    ElementID element_id;
    gint      offset;

};


static QuizItem*
quizitem_new (Card *card)
{
    QuizItem *q;

    q = g_new0 (QuizItem, 1);
    q->card = card;

    return q;
}


static GList*
shuffle_quizitems (GRand *grand, GList *list)
{
    GList *new = NULL;
    gint len, x;
    gpointer data;


    if (g_list_length (list) < 2) {
        return list;
    }

    while ((len = g_list_length (list)) > 0) {
        x = g_rand_int_range (grand, 0, len);
        data = g_list_nth_data (list, x);
        new = g_list_append (new, data);
        list = g_list_remove (list, data);
    }
    g_list_free (list);

    return new;
}


static GList*
quizitems_from_cards (GList *list, QuizInfo *info)
{
    GList *cur;
    GList *quizitems = NULL;
    gboolean add_it;


    for (cur = list; cur != NULL; cur = cur->next) {

        switch (info->card_selection) {
        case QUIZ_ALL_CARDS:
            add_it = TRUE;
            break;
        case QUIZ_NEW_CARDS:
            add_it = (card_get_group (CARD(cur)) == GROUP_NEW);
            break;
        case QUIZ_EXPIRED_CARDS:
            add_it = card_get_expired (CARD(cur));
            break;
        case QUIZ_NEW_AND_EXPIRED_CARDS:
            add_it = (card_get_group (CARD(cur)) == GROUP_NEW ||
                card_get_expired (CARD(cur)));
            break;
        case QUIZ_SELECTED_CARDS:
            add_it = TRUE;
            break;
        default:
            g_assert_not_reached ();
            break;
        }

        if (add_it) {

            quizitems = g_list_append (quizitems, quizitem_new (cur->data));

        }
    }
    return quizitems;
}


static GList*
get_quizitems (DialogData *data, GRand *grand)
{
    Category *cat;
    GList *cards = NULL;
    GList *quizitems;


    /* Build a list of cards to be tested. */

    if (data->ig->quizinfo.card_selection == QUIZ_SELECTED_CARDS) {

        cards = data->cardview_selected;

    }
    else {

        switch (data->ig->quizinfo.category_selection) {
        case QUIZ_CURRENT_CATEGORY:
            cat = file_get_current_category (data->ig->file);
            if (category_get_n_cards (cat) > 0) {
                cards = category_get_cards (cat);
            }
            break;
        case QUIZ_ALL_CATEGORIES:
            cards = file_get_cards (data->ig->file);
            break;
        default:
            g_assert_not_reached ();
            break;
        }

    }

    if (!cards) { return NULL; }

    quizitems = quizitems_from_cards (cards, &data->ig->quizinfo);
    quizitems = shuffle_quizitems (grand, quizitems);

    return quizitems;
}


static void
store_quiz_results (DialogData *data, gboolean store_statistics)
{
    GList *cur;


    /* Update the statistics of each tested card. */

    for (cur = data->quizitems; cur != NULL; cur = cur->next) {

        if (QUIZITEM(cur)->tested) {

            if (store_statistics) {

                card_mark_tested (data->ig->prefs, QCARD(cur),
                    QUIZITEM(cur)->known);

            }

            /* Move unknown cards to the top in their categories. */
            if (!QUIZITEM(cur)->known) {
                card_to_top (QCARD(cur));
            }

            file_set_changed (data->ig->file, TRUE);

        }
    }

    file_check_expired (data->ig->file);

    app_window_refresh_card_pane (data->ig,
        file_get_current_category_cards (data->ig->file));

    /* Add an asterisk to the main window title if the file has changed. */
    app_window_update_title (data->ig);
}


static void
statusbar_set_message (DialogData *data, const gchar *text)
{
    gtk_statusbar_pop (data->statusbar, data->sbcid);
    gtk_statusbar_push (data->statusbar, data->sbcid, text);
}


static void
textbuf_clear (GtkTextBuffer *textbuf)
{
    GtkTextIter start, end;

    gtk_text_buffer_get_start_iter (textbuf, &start);
    gtk_text_buffer_get_end_iter (textbuf, &end);
    gtk_text_buffer_delete (textbuf, &start, &end);
}


static void
textbuf_put (GtkTextBuffer *textbuf, const gchar *text)
{
    GtkTextIter iter;

    g_assert (text != NULL);
    gtk_text_buffer_get_end_iter (textbuf, &iter);
    gtk_text_buffer_insert (textbuf, &iter, text, -1);
}


static Parser*
parser_new (void)
{
    Parser *p;

    p = g_new0 (Parser, 1);
    p->data = g_string_new ("");

    return p;
}


static void
parser_free (Parser *p)
{
    g_string_free (p->data, TRUE);
    if (p->tags != NULL) {
        g_list_foreach (p->tags, (GFunc)g_free, NULL);
        g_list_free (p->tags);
    }
    g_free (p);
}


static ElementID
parser_parse_tag (const gchar *str, gboolean *is_open_tag)
{
    gint id, length;

    str++;
    if (*str == '/') {
        *is_open_tag = FALSE;
        str++;
    }
    else {
        *is_open_tag = TRUE;
    }

    if (*str == '\0') {
        return ELEMENT_ID_UNKNOWN;
    }

    for (id = 1; id < n_elements; id++) {

        length = element[id].length;

        if (strncmp (str, element[id].name, length) == 0) {
            if (str[length] == '>') {
                return id;
            }
        }

    }

    return ELEMENT_ID_UNKNOWN;
}


static Token
parser_get_token (const gchar *text, Parser *p)
{
    const gchar *s, *s_start;
    ElementID id;


    s = s_start = text;
    g_string_set_size (p->data, 0);

    p->length = 0;

    if (*s == '\0') {
        if (p->in_keyword) {
            p->token = TOKEN_NEW_LINE;
        }
        else {
            p->token = TOKEN_NONE;
        }
        return p->token;
    }

    if (*s == '<' && !p->error) {

        id = parser_parse_tag (s, &p->is_open_tag);

        if (id != ELEMENT_ID_UNKNOWN) {

            if (p->is_open_tag) {
                p->length = element[id].length + 2;
            }
            else {
                p->length = element[id].length + 3;
            }
            p->element_id = id;
            p->token = TOKEN_TAG;

            return p->token;
        }

    }

    if (*s == '\n') {

        p->length = 1;
        p->token = TOKEN_NEW_LINE;

        g_string_assign (p->data, "\n");

        return p->token;
    }

    while (*s != '\0') {

        if (*s == '<' && s != s_start) {
            break;
        }
        if (*s == '\n') {
            break;
        }

        s++;

        p->length = p->length + 1;
    }

    g_string_overwrite_len (p->data, 0, s_start, p->length);

    p->token = TOKEN_TEXT;

    return p->token;
}


static void
display_open_tag (GtkTextBuffer *textbuf, Parser *p)
{
    GtkTextIter start;
    Tag *tag;

    gtk_text_buffer_get_end_iter (textbuf, &start);

    tag = g_new0 (Tag, 1);
    tag->element_id = p->element_id;
    tag->offset = gtk_text_iter_get_offset (&start);

    p->tags = g_list_prepend (p->tags, tag);
}


static void
display_close_tag (GtkTextBuffer *textbuf, Parser *p)
{
    GtkTextIter start, end;
    Tag *tag;

    if ((tag = g_list_first (p->tags)->data) != NULL) {

        if (tag->element_id != p->element_id) {
            g_warning ("Tag mismatch\n");
            p->error = TRUE;
            return;
        }

        gtk_text_buffer_get_iter_at_offset (textbuf, &start, tag->offset);
        gtk_text_buffer_get_end_iter (textbuf, &end);

        gtk_text_buffer_apply_tag_by_name (textbuf,
            element[tag->element_id].name, &start, &end);

        p->tags = g_list_delete_link (p->tags, g_list_first (p->tags));
    }
}


static void
display_text (GtkTextBuffer *textbuf, const gchar *text, CardStyle style)
{
    Parser *p;
    const gchar *s;


    if (text == NULL) { text = ""; }

    textbuf_clear (textbuf);

    p = parser_new ();

    if (style != CARD_STYLE_SENTENCES) {
        p->in_keyword = TRUE;
        p->element_id = ELEMENT_ID_KEYWORD;
        display_open_tag (textbuf, p);
    }

    s = text;

    while (parser_get_token (s, p) != TOKEN_NONE) {
        switch (p->token) {
        case TOKEN_TEXT:
            textbuf_put (textbuf, p->data->str);
            break;
        case TOKEN_NEW_LINE:
            if (p->in_keyword) {
                p->in_keyword = FALSE;
                p->element_id = ELEMENT_ID_KEYWORD;
                display_close_tag (textbuf, p);
            }
            textbuf_put (textbuf, p->data->str);
            break;
        case TOKEN_TAG:
            if (p->is_open_tag) {
                display_open_tag (textbuf, p);
            }
            else {
                display_close_tag (textbuf, p);
            }
            break;
        default:
            break;
        }
        s = s + p->length;
    }

    parser_free (p);
}


gchar*
remove_card_markup (const gchar *text)
{
    Parser *p;
    const gchar *s;
    gchar *plain_text;
    GString *gstr;

    p = parser_new ();

    gstr = g_string_new ("");

    s = text;

    while (parser_get_token (s, p) != TOKEN_NONE) {
        switch (p->token) {
        case TOKEN_TEXT:
        case TOKEN_NEW_LINE:
            g_string_append (gstr, p->data->str);
            break;
        default:
            break;
        }
        s = s + p->length;
    }

    parser_free (p);

    plain_text = gstr->str;
    g_string_free (gstr, FALSE);

    return plain_text;
}


static void
show_current_card (DialogData *data)
{
    const gchar *text;
    gboolean flip, active;
    Card *c;
    GdkColor *color;
    CardStyle style;


    c = QCARD(data->cur);

    /* Set colours and statusbar message. */

    if (CUR_IS_LAST(data)) {

        color = prefs_get_color (data->ig->prefs, COLOR_CARD_BG_END);
        statusbar_set_message (data, "");
        gtk_widget_grab_focus (data->b_prev);

    }
    else {

        gchar *s;

        color = prefs_get_color (data->ig->prefs, COLOR_CARD_BG);
        s = g_strdup_printf ("Card %d of %d. Category: %s",
            data->n,
            data->n_quizitems,
            category_get_title (card_get_category (c)));
        statusbar_set_message (data, s);
        g_free (s);

    }
    gtk_widget_modify_base (GTK_WIDGET(data->textview),
        GTK_STATE_NORMAL, color);


    /* Set state of known/unknown buttons. */

    data->ugly_toggle_hack = TRUE;
    active = QUIZITEM(data->cur)->tested && !QUIZITEM(data->cur)->known;
    gtk_toggle_button_set_active (data->b_unknown, active);
    gtk_check_menu_item_set_active (data->m_unknown, active);
    if (active) {

        /* Highlight cards marked as unknown while we're at it. */
        color = prefs_get_color (data->ig->prefs, COLOR_CARD_BG_UNKNOWN);
        gtk_widget_modify_base (GTK_WIDGET(data->textview),
            GTK_STATE_NORMAL, color);

    }
    else if (QUIZITEM(data->cur)->tested) {

        color = prefs_get_color (data->ig->prefs, COLOR_CARD_BG_KNOWN);
        gtk_widget_modify_base (GTK_WIDGET(data->textview),
            GTK_STATE_NORMAL, color);

    }

    active = QUIZITEM(data->cur)->tested && QUIZITEM(data->cur)->known;
    gtk_toggle_button_set_active (data->b_known, active);
    gtk_check_menu_item_set_active (data->m_known, active);
    data->ugly_toggle_hack = FALSE;


    /* Display card text. */

    flip = gtk_toggle_button_get_active (data->b_flip);
    if (data->ig->quizinfo.reverse && !CUR_IS_LAST(data)) {
        flip = !flip;
    }
    text = flip ? card_get_back (c) : card_get_front (c);

    style = file_card_get_card_style (data->ig->file, c);
    display_text (data->textbuf, text, style);
}


static void
set_textview_properties (DialogData *data)
{
    GdkColor *color;
    const gchar *font;
    PangoFontDescription *desc;
    GtkJustification just_keyword, just_body;
    CardStyle style;


    style = file_get_card_style (data->ig->file);

    switch (style) {
    case CARD_STYLE_NONE:
    case CARD_STYLE_KEYWORDS_AND_CENTERED_TEXT:
        just_keyword = GTK_JUSTIFY_CENTER;
        just_body = GTK_JUSTIFY_CENTER;
        break;
    case CARD_STYLE_KEYWORDS_AND_SENTENCES:
    case CARD_STYLE_SENTENCES:
        just_keyword = GTK_JUSTIFY_LEFT;
        just_body = GTK_JUSTIFY_LEFT;
        break;
    }
    gtk_text_view_set_justification (GTK_TEXT_VIEW(data->textview), just_body);

    gtk_text_buffer_create_tag (data->textbuf, "i",
        "style", PANGO_STYLE_ITALIC,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "b",
        "weight", PANGO_WEIGHT_BOLD,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "u",
        "underline", PANGO_UNDERLINE_SINGLE,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "s",
        "strikethrough", TRUE,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "big",
        "scale", PANGO_SCALE_X_LARGE,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "small",
        "scale", PANGO_SCALE_X_SMALL,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "sup",
        "scale", PANGO_SCALE_X_SMALL,
        "rise", 5000,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "sub",
        "scale", PANGO_SCALE_X_SMALL,
        "rise", -4000,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "key",
        "scale", PANGO_SCALE_XX_LARGE,
        "justification", just_keyword,
        NULL);
    gtk_text_buffer_create_tag (data->textbuf, "center",
        "justification", GTK_JUSTIFY_CENTER,
        NULL);

    color = prefs_get_color (data->ig->prefs, COLOR_CARD_FG);
    gtk_widget_modify_text (GTK_WIDGET(data->textview),
        GTK_STATE_NORMAL, color);

    if ((font = prefs_get_card_font (data->ig->prefs)) != NULL) {
        desc = pango_font_description_from_string (font);
        gtk_widget_modify_font (GTK_WIDGET(data->textview), desc);
        pango_font_description_free (desc);
    }

    gtk_text_view_set_left_margin (data->textview, 6);
    gtk_text_view_set_right_margin (data->textview, 6);

    gtk_text_view_set_editable (data->textview, FALSE);
    gtk_text_view_set_cursor_visible (data->textview, FALSE);
}


static void
set_interface_sensitivity (DialogData *data)
{
    gboolean last;

    last = CUR_IS_LAST(data);

    gtk_widget_set_sensitive (data->b_prev, data->cur->prev != NULL);
    gtk_widget_set_sensitive (data->b_next, data->cur->next != NULL);

    gtk_widget_set_sensitive (data->m_prev, data->cur->prev != NULL);
    gtk_widget_set_sensitive (data->m_next, data->cur->next != NULL);

    gtk_widget_set_sensitive (GTK_WIDGET(data->b_unknown), !last);
    gtk_widget_set_sensitive (GTK_WIDGET(data->b_known), !last);
    gtk_widget_set_sensitive (GTK_WIDGET(data->b_flip), !last);

    gtk_widget_set_sensitive (GTK_WIDGET(data->m_unknown), !last);
    gtk_widget_set_sensitive (GTK_WIDGET(data->m_known), !last);
    gtk_widget_set_sensitive (GTK_WIDGET(data->m_flip), !last);

    gtk_widget_set_sensitive (GTK_WIDGET(data->chk_store),
        data->n_quizitems > 0);
}


static gboolean
step_forward (DialogData *data)
{
    /* Switch to the next card, or return FALSE. */

    if (data->cur->next) {

        data->cur = data->cur->next;
        data->n = data->n + 1;

        gtk_toggle_button_set_active (data->b_flip, FALSE); /* Un-flip */

        set_interface_sensitivity (data);

        return TRUE;

    }
    return FALSE;
}


static gboolean
step_back (DialogData *data)
{
    /* Switch to the previous card, or return FALSE. */

    if (data->cur->prev) {

        data->cur = data->cur->prev;
        data->n = data->n - 1;

        gtk_toggle_button_set_active (data->b_flip, FALSE); /* Un-flip */

        if (data->cur->prev == NULL) {
            /* Can't go any further back, so focus the "Next" button. */
            gtk_widget_grab_focus (data->b_next);
        }

        set_interface_sensitivity (data);

        return TRUE;

    }
    return FALSE;
}


static void
cb_m_prev (GtkWidget *widget, DialogData *data)
{
    step_back (data);
    show_current_card (data);
}


static void
cb_m_next (GtkWidget *widget, DialogData *data)
{
    step_forward (data);
    show_current_card (data);
}


static void
cb_b_unknown (GtkToggleButton *tb, DialogData *data)
{
    gboolean marked_as_tested;

    if (data->ugly_toggle_hack == TRUE) { return; }

    if (!gtk_toggle_button_get_active (tb) &&
        !gtk_toggle_button_get_active (data->b_known)) {
        QUIZITEM(data->cur)->tested = FALSE;
        show_current_card (data);
        return;
    }

    marked_as_tested = QUIZITEM(data->cur)->tested;

    gtk_toggle_button_set_active (data->b_known, FALSE);

    QUIZITEM(data->cur)->known = FALSE;
    QUIZITEM(data->cur)->tested = TRUE;

    /* If the card was marked as untested, move on to the next card. */

    if (!marked_as_tested) {
        step_forward (data);
    }

    show_current_card (data);
}


static void
cb_m_unknown (GtkCheckMenuItem *widget, DialogData *data)
{
    gtk_toggle_button_set_active (data->b_unknown,
        gtk_check_menu_item_get_active (widget));
}


static void
cb_b_known (GtkToggleButton *tb, DialogData *data)
{
    gboolean marked_as_tested;

    if (data->ugly_toggle_hack == TRUE) { return; }

    if (!gtk_toggle_button_get_active (tb) &&
        !gtk_toggle_button_get_active (data->b_unknown)) {
        QUIZITEM(data->cur)->tested = FALSE;
        show_current_card (data);
        return;
    }

    marked_as_tested = QUIZITEM(data->cur)->tested;

    gtk_toggle_button_set_active (data->b_unknown, FALSE);

    QUIZITEM(data->cur)->known = TRUE;
    QUIZITEM(data->cur)->tested = TRUE;

    /* If the card was marked as untested, move on to the next card. */

    if (!marked_as_tested) {
        step_forward (data);
    }

    show_current_card (data);
}


static void
cb_m_known (GtkCheckMenuItem *widget, DialogData *data)
{
    gtk_toggle_button_set_active (data->b_known,
        gtk_check_menu_item_get_active (widget));
}


static void
cb_b_flip (GtkToggleButton *tb, DialogData *data)
{
    gtk_check_menu_item_set_active (data->m_flip,
        gtk_toggle_button_get_active (tb)); 

    show_current_card (data);
}


static void
cb_m_flip (GtkCheckMenuItem *widget, DialogData *data)
{
    gtk_toggle_button_set_active (data->b_flip,
        gtk_check_menu_item_get_active (widget)); 
}


static void
window_close (DialogData *data)
{
    gint w, h;
    GList *cur;
    QuizItem *last;

    store_quiz_results (data, gtk_toggle_button_get_active
        (GTK_TOGGLE_BUTTON(data->chk_store)));

    last = data->last;
    card_free (last->card);

    for (cur = data->quizitems; cur != NULL; cur = cur->next) {
        g_free (cur->data);
    }
    g_list_free (data->quizitems);

    if (data->cardview_selected) {
        g_list_free (data->cardview_selected);
    }
    /* g_free (data->info); */

    gtk_window_get_size (GTK_WINDOW(data->window), &w, &h);
    prefs_set_quiz_size (data->ig->prefs, w, h);

    gtk_widget_destroy (data->window);

    g_object_unref (G_OBJECT(data->gui));
    g_free (data);
}


static void
cb_m_close (GtkWidget *widget, DialogData *data)
{
    window_close (data);
}


static void
cb_window_delete (GtkWidget *widget, GdkEvent *event, DialogData *data)
{
    window_close (data);
}


static gboolean
add_to_card_selection (GtkTreeModel  *model,
    GtkTreePath *path, GtkTreeIter *iter, DialogData *data)
{
    GList *item;

    gtk_tree_model_get (model, iter, COLUMN_CARD_DATA, &item, -1);
    data->cardview_selected = g_list_append (data->cardview_selected,
        CARD(item));
    return TRUE;
}

void
dialog_quiz (Ignuit *ig)
{
    DialogData *dialog;
    GtkWidget *m_close, *img;
    Card *c;
    gchar *gladefile, *imgfile;


    gladefile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, F_GLADE_QUIZ, TRUE, NULL);

    if (gladefile == NULL) {
        g_warning ("Can't find file: %s\n", F_GLADE_QUIZ);
        return;
    }

    dialog = g_new0 (DialogData, 1);

    dialog->gui = glade_xml_new (gladefile, NULL, NULL);
    g_free (gladefile);

    dialog->ig = ig;

    dialog->cardview_selected = NULL;
 
    if (dialog->ig->quizinfo.card_selection == QUIZ_SELECTED_CARDS) {

        /* Get a list of selected items from the main window's card treeview. */

        GtkTreeSelection *sel;
        GtkTreeModel *model;
        GList *rows;

        sel = gtk_tree_view_get_selection (dialog->ig->treev_card);
        rows = gtk_tree_selection_get_selected_rows (sel, &model);
        if (rows) {

            gtk_tree_selection_selected_foreach (sel,
                (GtkTreeSelectionForeachFunc)add_to_card_selection, dialog);

            g_list_foreach (rows, (GFunc)gtk_tree_path_free, NULL);
            g_list_free (rows);
        }

    }

    dialog->quizitems = get_quizitems (dialog, ig->grand);

    c = card_new ();
    dialog->last = quizitem_new (c);
    dialog->quizitems = g_list_append (dialog->quizitems, dialog->last);
    dialog->cur = dialog->quizitems;

    card_set_style (c, CARD_STYLE_KEYWORDS_AND_CENTERED_TEXT);

    dialog->n = 1;
    dialog->n_quizitems = g_list_length (dialog->quizitems) - 1;

    if (dialog->n_quizitems == 0) {

        switch (dialog->ig->quizinfo.card_selection) {
        case QUIZ_ALL_CARDS:
            /* GUI sensitivity should've prevented us from starting
             * a quiz in an empty file or empty category. */
            g_assert_not_reached ();
            break;
        case QUIZ_NEW_CARDS:
            card_set_front (c,
            _("<center>No cards.</center>\n\nThere are no new cards to test.\nSee the <i>Quiz</i> menu for options."));
            break;
        case QUIZ_EXPIRED_CARDS:
            card_set_front (c,
            _("<center>No cards.</center>\n\nThere are no expired cards to test.\nSee the <i>Quiz</i> menu for options."));
            break;
        case QUIZ_NEW_AND_EXPIRED_CARDS:
            card_set_front (c,
            _("<center>No cards.</center>\n\nThere are no new or expired cards to test.\nSee the <i>Quiz</i> menu for options."));
            break;
        case QUIZ_SELECTED_CARDS:
            card_set_front (c,
            _("<center>No cards.</center>\n\nThere are no cards selected.\nSee the <i>Quiz</i> menu for options."));
            break;
        default:
            g_assert_not_reached ();
            break;
        }

    }
    else {
        card_set_front (c,
            _("<center>End of quiz.</center>\n\nUse the buttons below to review your choices. Only cards marked as <i>Known</i> or <i>Unknown</i> are recorded. Nothing is recorded until the quiz window closes."));
    }


    dialog->window = glade_xml_get_widget (dialog->gui, "dialog");

    gtk_window_set_default_size (GTK_WINDOW(dialog->window),
        prefs_get_quiz_width (dialog->ig->prefs),
        prefs_get_quiz_height (dialog->ig->prefs));

    dialog->textview = GTK_TEXT_VIEW(glade_xml_get_widget (dialog->gui, "textview"));
    dialog->textbuf = gtk_text_view_get_buffer (dialog->textview);

    dialog->m_prev = glade_xml_get_widget (dialog->gui, "m_prev");
    dialog->m_next = glade_xml_get_widget (dialog->gui, "m_next");
    dialog->m_unknown = (GtkCheckMenuItem*)glade_xml_get_widget
        (dialog->gui, "m_unknown");
    dialog->m_known = (GtkCheckMenuItem*)glade_xml_get_widget
        (dialog->gui, "m_known");
    dialog->m_flip = (GtkCheckMenuItem*)glade_xml_get_widget
        (dialog->gui, "m_flip");
    m_close = glade_xml_get_widget (dialog->gui, "m_close");

    dialog->b_prev = glade_xml_get_widget (dialog->gui, "b_prev");
    dialog->b_next = glade_xml_get_widget (dialog->gui, "b_next");
    dialog->chk_store = glade_xml_get_widget (dialog->gui, "chk_store");
    dialog->b_unknown = (GtkToggleButton*)glade_xml_get_widget
        (dialog->gui, "b_unknown");
    dialog->b_known = (GtkToggleButton*)glade_xml_get_widget
        (dialog->gui, "b_known");
    dialog->b_flip = (GtkToggleButton*)glade_xml_get_widget
        (dialog->gui, "b_flip");

    dialog->statusbar = (GtkStatusbar*)glade_xml_get_widget
        (dialog->gui, "statusbar");
    dialog->sbcid = gtk_statusbar_get_context_id (dialog->statusbar, "quiz");

    if ((imgfile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, F_IMG_BTN_UNKNOWN, TRUE, NULL))) {
        img = gtk_image_new_from_file (imgfile);
        gtk_button_set_image (GTK_BUTTON(dialog->b_unknown), img);
        g_free (imgfile);
        gtk_widget_show (img);
    }
    else {
        gtk_button_set_label (GTK_BUTTON(dialog->b_unknown), "Unknown");
    }

    if ((imgfile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, F_IMG_BTN_KNOWN, TRUE, NULL))) {
        img = gtk_image_new_from_file (imgfile);
        gtk_button_set_image (GTK_BUTTON(dialog->b_known), img);
        g_free (imgfile);
        gtk_widget_show (img);
    }
    else {
        gtk_button_set_label (GTK_BUTTON(dialog->b_known), "Known");
    }

#if 0
    if ((imgfile = gnome_program_locate_file (ig->program,
        GNOME_FILE_DOMAIN_APP_DATADIR, F_IMG_BTN_FLIP, TRUE, NULL))) {
        img = gtk_image_new_from_file (imgfile);
        gtk_button_set_image (GTK_BUTTON(dialog->b_flip), img);
        g_free (imgfile);
        gtk_widget_show (img);
    }
    else {
        gtk_button_set_label (GTK_BUTTON(dialog->b_flip), "Flip");
    }
#endif

    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(dialog->chk_store),
        dialog->ig->quizinfo.card_selection == QUIZ_NEW_AND_EXPIRED_CARDS ||
        dialog->ig->quizinfo.card_selection == QUIZ_EXPIRED_CARDS);

    g_signal_connect (G_OBJECT(dialog->window), "delete-event",
        G_CALLBACK(cb_window_delete), dialog);

    g_signal_connect (G_OBJECT(dialog->m_prev), "activate",
        G_CALLBACK(cb_m_prev), dialog);
    g_signal_connect (G_OBJECT(dialog->m_next), "activate",
        G_CALLBACK(cb_m_next), dialog);
    g_signal_connect (G_OBJECT(dialog->m_unknown), "toggled",
        G_CALLBACK(cb_m_unknown), dialog);
    g_signal_connect (G_OBJECT(dialog->m_known), "toggled",
        G_CALLBACK(cb_m_known), dialog);

    g_signal_connect (G_OBJECT(dialog->m_flip), "toggled",
        G_CALLBACK(cb_m_flip), dialog);
    g_signal_connect (G_OBJECT(m_close), "activate",
        G_CALLBACK(cb_m_close), dialog);

    g_signal_connect (G_OBJECT(dialog->b_prev), "clicked",
        G_CALLBACK(cb_m_prev), dialog);
    g_signal_connect (G_OBJECT(dialog->b_next), "clicked",
        G_CALLBACK(cb_m_next), dialog);
    g_signal_connect (G_OBJECT(dialog->b_unknown), "toggled",
        G_CALLBACK(cb_b_unknown), dialog);
    g_signal_connect (G_OBJECT(dialog->b_known), "toggled",
        G_CALLBACK(cb_b_known), dialog);
    g_signal_connect (G_OBJECT(dialog->b_flip), "toggled",
        G_CALLBACK(cb_b_flip), dialog);

    gtk_window_set_transient_for (GTK_WINDOW(dialog->window),
        GTK_WINDOW(ig->app));
    gtk_window_set_modal (GTK_WINDOW(dialog->window), TRUE);

    set_textview_properties (dialog);
    show_current_card (dialog);
    set_interface_sensitivity (dialog);

    gtk_widget_show (dialog->window);
}

